# trigger.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

All trigger functions.

"""

from mma.MMA.common import *
import mma.MMA.debug
from . import gbl
import mma.MMA.truncate

import random
import copy 

# Each track class has a pointer to a unique copy of this object
class Trigger:
    def __init__(self):
        self.mode = None
        self.truncate = False
        self.count = 1
        self.bars = []
        self.beats = []
        self.seq = None
        self.measures = []
        self.cnames = []
        self.ctonics = []
        self.ctypes = []
        self.override = False


def makeTriggerSequence(self, ctable, pattern):
    """ Create a new sequence based on a trigger setting. 
        This is called only from bar().
    """

    trigger = self.trigger
    
    if trigger.bars and gbl.seqCount not in trigger.bars:
        return []

    if trigger.measures and gbl.barLabel not in trigger.measures:
        return []

    # Use the internal trigger sequence if it's set.
    # otherwise, use the pattern(s) from the
    # sequence set for the track.
    if trigger.seq:
        pattern = trigger.seq

    # Not redundant! 
    if not pattern:
        return []
    
    tpats = []

    # If we have trigger set to "AUTO" then we create a set of offsets
    # based on the chord changes in ctable.
    if trigger.mode == 'AUTO':
        trigs = []
        for c in ctable:
            if c.name != 'z' and c.name != c.lastchord:
                trigs.append(c.chStart)

    # This works with 'z' chords only, not 'XXzYY'
    elif trigger.mode == 'REST':
        trigs = []
        for c in ctable:
            if c.name == 'z' and c.name != c.lastchord:
                trigs.append(c.chStart)                    

    # different 'beat' combinations
    elif trigger.mode == 'BEATS':

        # If chord names are set set if the beat and the
        # chord NAME at that point
        if trigger.cnames:
            trigs = []
            for t in trigger.beats:
                for c in ctable:
                    if t >= c.chStart and t < c.chEnd and \
                             c.chord.name in trigger.cnames:
                        trigs.append(t)
                        break

        # Chord tonic names (C, D, E ... G)
        elif trigger.ctonics:
            trigs = []
            for t in trigger.beats:
                for c in ctable:
                    if t >= c.chStart and t < c.chEnd and \
                            c.chord.tonic in trigger.ctonics:
                        trigs.append(t)
                        break

        elif trigger.ctypes:
            trigs = []
            for t in trigger.beats:
                for c in ctable:
                    if t >= c.chStart and t < c.chEnd and \
                           c.chord.chordType in trigger.ctypes:
                        trigs.append(t)
                        break
                    
        # use the user defined beat list
        else:
            trigs = trigger.beats
        
    else:
        return []

    # make copies of pattern(s) for each beat.
    # the number of patterns to copy are determined by the 'count'
    # option (default==1).
    
    c = min(len(pattern), trigger.count)
    if c < 1:  # either no pattern or no count
        return []

    for t in trigs:
        for i in range(c):
            base = copy.deepcopy(pattern[i])
            base.offset += t 
            if base.offset >= gbl.barLen:
                continue
            tpats.append(base)

    # normalize the durations of the new patterns so they
    # terminate at the next pattern or the end of the bar
    if trigger.truncate:
        for i in range(len(tpats)):
           if i < len(tpats) - 1:  # another chord follows, truncate at start of next chord
               maxd = tpats[i+1].offset - tpats[i].offset
           else: # last chord, truncate at end of bar
               # Note to braindead self, the TRUNCATE option is
               # used to shorten a bar's duration. So the end of
               # bar can be 'gbl.barlen' or 'truncate.length'
               if mma.MMA.truncate.length:
                   barEnd = mma.MMA.truncate.length
               else:
                   barEnd = gbl.barLen
               maxd = barEnd - tpats[i].offset
           tpats[i].duration = min(tpats[i].duration, maxd)
    return tpats


def setTrigger(name, ln):
    """ Set a trigger for a track. 

        A trigger is simply a list of offsets which are processed
        by the trackBar() functions for each track.

        This is called from the parser. We grab the track class from 'name'.

    """

    self = gbl.tnames[name]

    if self.vtype in ('MELODY', 'SOLO'):
        error("Trigger is not valid in %s track." % self.vtype)

    self.trigger = trigger = Trigger()  # always reset to default

    sequence = None

    if not ln:        # empty line is same as singleton 'off'
        ln = ['OFF']

    argCount = len(ln)
    trigger.mode = 'ON'

    # extract a possible sequence. Only 1st is extracted and
    # saved for use by the 'SEQUENCE = ' command. The stuff in {}
    # is replaced by an empty {}.
    l = ' '.join(ln)
    if '{' in l and '}' in l:
        ln, sequence = pextract(l, '{', '}', onlyone=True, insert='{}')
        ln = ln.split()
        sequence = sequence[0]

    ln, opts = opt2pair(ln)

    ##########################
    # do opts with args

    for cmd, opt in opts:
        cmd = cmd.upper()
        if cmd == 'BEATS':
            trigger.mode = 'BEATS'
            for t in opt.split(','):
                trigger.beats.append(self.setBarOffset(t))
            trigger.beats.sort()

        elif cmd == 'BARS':
            for t in opt.split(','):
                v = stoi(t)
                if v < 1 or v > gbl.seqSize:
                    warning("%s Trigger Bars: setting of %s may be "
                            "ignored since SeqSize is %s." %
                            (self.name, v, gbl.seqSize))
                trigger.bars.append(v - 1)

        elif cmd == 'CNAMES':
            for a in opt.split(','):
                trigger.cnames.append(a)

        elif cmd == 'CTYPES':
            for a in opt.split(','):
                trigger.ctypes.append(a)

        elif cmd == 'CTONICS':
            for a in opt.split(','):
                trigger.ctonics.append(a)

        elif cmd == 'COUNT':
            trigger.count = stoi(opt)
            if trigger.count <= 0:
                error("%s Trigger Count must be greater than 0, not '%s'." %
                      (self.name, trigger.count))

        elif cmd == 'MEASURES':
            for a in opt.split(','):
                if not a.isdigit():
                    warning("%s Trigger: Measures should be bar labels, not '%s'."
                            "This trigger will be ignored." % (self.name, a))
                    continue
                trigger.measures.append(a.lstrip('0'))

        elif cmd == 'OVERRIDE':
            if opt.upper() in ('ON', 1, 'TRUE'):
                trigger.override = True
            elif opt.upper() in ('OFF', 0, 'FALSE'):
                trigger.override = False

        elif cmd == 'TRUNCATE':
            if opt.upper() in ('ON', 1, 'TRUE'):
                trigger.truncate = True
            elif opt.upper() in ('OFF', 0, 'FALSE'):
                trigger.truncate = False
            else:
                error("%s Trigger Truncate expecting ON OFF, not %s." %
                      (self.name, opt))

        elif cmd == 'SEQUENCE':
            if sequence:
                sequence = sequence.rstrip('; ')
                trigger.seq = self.defPatRiff(sequence)
            else:
                error("%s Trigger Sequence expecting {patterns...}." % self.name)

        elif cmd == 'STICKY':
            if opt.upper() in ('ON', 1, 'TRUE'):
                self.sticky = True
            elif opt.upper() in ('OFF', 0, 'FALSE'):
                self.sticky == False
            else:
                error("%s Trigger Sticky expecting ON OFF, not %s" % \
                      (self.name, opt))

        else:
            error("%s Trigger '%s' is an unknown command." % (self.name, cmd))

    # opts without args
    for cmd in ln:
        cmd = cmd.upper()
        if cmd == 'AUTO':
            trigger.mode = 'AUTO'

        elif cmd == 'OFF': # It's all been set to default at top, so leave it alone
            if argCount > 1:
                error("Trigger: 'OFF' argument must be the only arg.")

        elif cmd == 'REST':
            trigger.mode = 'REST'

        else:
            error("%s Trigger '%s' is an unknown command." % (self.name, cmd))

    if mma.MMA.debug.debug:
        mma.MMA.debug.trackSet(self.name, "TRIGGER")

def getTriggerOptions(self):
    """ Called from setTrigger() and macro. Returns string with current options. """

    trigger = self.trigger

    mode = ''

    if trigger.mode in ('AUTO', 'REST'):
        mode = trigger.mode

    
    if not trigger.beats:
        beats = '[]'
    else:
        beats =  ','.join([str(1 + (i / float(gbl.BperQ))) for i in trigger.beats])

    if trigger.cnames:
        cnames = ','.join([i for i in trigger.cnames])
    else:
        cnames = '[]'

    if trigger.ctypes: 
        ctypes = ','.join([i for i in trigger.ctypes])
    else:
        ctypes = '[]'

    if trigger.ctonics:
        tonics = ','.join([i for i in trigger.ctonics])
    else:
        tonics = '[]'

    if not trigger.bars:
        bars = '[]'
    else:
        bars = ','.join([str(a + 1) for a in trigger.bars])

    if trigger.seq:
        seq =  "Sequence={%s}" % self.formatPattern(trigger.seq)
    else:
        seq = "Sequence={}"

    return "%s Beats=%s CNames=%s CTypes=%s CTonics=%s Bars=%s" \
              "Count=%s Truncate=%s Override=%s %s" % \
          ( mode, beats, cnames, ctypes, tonics, bars, trigger.count,
            trigger.truncate, trigger.override, seq)





# ornament.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

All ornamentation functions.

"""

from mma.MMA.common import *
from . import gbl
import mma.MMA.notelen
import random
import mma.MMA.debug

def default():
    return {'type': None, 'chromatic': False, 'duration': .2, 'pad': (.1, .1),
            'place': 'ABOVE', 'volume': .75, 'beats': (), 'bars': (), 'rskip': 0,
            'rvolume': (0,0), 'count': 5, 'offset': 0}


def getOrnOpts(self):
    """ Return the ornament settings. Used by debug and macro code. """

    o = self.ornaments

    if o['bars']:
        bars = ','.join([str(i + 1) for i in o['bars']])
    else:
        bars = 'ALL'
    if o['beats']:
        beats = ','.join([str(1 + (i / float(gbl.BperQ))) for i in o['beats']])
    else:
        beats = 'ALL'

    return "Type=%s Chromatic=%s Duration=%s Count=%s Pad=%s,%s Offset=%s Volume=%s " \
      "RVolume=%s,%s Place=%s Beats=%s Rskip=%s Bars=%s" % \
      (str(o['type']).upper(),
       ('OFF', 'ON')[o['chromatic']],
       o['duration'] * 100,
       o['count'],
       o['pad'][0] * 100, o['pad'][1] * 100,
       o['offset'],
       o['volume'] * 100,
       o['rvolume'][0] * 100, o['rvolume'][1] * 100,
       o['place'],
       beats,
       o['rskip'] * 100,
       bars)


def setOrnament(self, ln):
    """ Add an ornamentation option. """

    o = self.ornaments = default()
    notopt, optpair = opt2pair(ln)

    if self.vtype not in ("CHORD", "BASS", "WALK", "ARPEGGIO", "SCALE"):
        error("Ornament is not valid in %s tracks." % self.vtype)

    if (notopt and len(notopt) == 1 and notopt[0].upper() in ('NONE', 'OFF')) \
            or (not notopt and not optpair):
        o['type'] = None
        if mma.MMA.debug.debug:
            dPrint("%s Ornament: Off" % self.name)
        return

    if notopt or not optpair:
        error("%s Ornament: expecting cmd=opt pairs,"
              "not '%s'." % (self.name, ' '.join(notopt)))

    for cmd, opt in optpair:
        cmd = cmd.upper()
        opt = opt.upper()
        
        if cmd == 'TYPE':
            if opt in ('MORDENT', 'TRILL', 'TURN', 'GRACE',
                       '3AFTER', 'FALL', 'GLISS', 'GLISSAFTER' ):
                o['type'] = opt
            else:
                error("%s Ornament Type: '%s' is an unknown type." % (self.name, opt))

        elif cmd == 'COUNT':
            o['count'] = stoi(opt)
            if o['count'] <= 0:
                error("%s Ornament Count: value must be positive integer, not '%s'." % 
                      (self.name, opt))

        elif cmd == 'CHROMATIC':
            if opt in ('ON', 'TRUE'):
                o['chromatic'] = True
            elif opt in ('OFF', 'FALSE'):
                o['chromatic'] = False
            else:
                error("%s Ornament Chromatic: value must be 'On' or 'Off', not '%s'." %
                      (self.name, opt))

        elif cmd == 'DURATION':
            v = stof(opt)
            if v < 0 or v > 100:
                error("%s Ornament Duration: values must be greater "
                      "than 0 and less that 100, not '%s'." % (self.name, v))
            o['duration'] = v / 100

        elif cmd == 'PAD':
            o['pad'] = []
            for v in opt.split(',', 1):
                v = stof(v)
                if v < -100 or v > 100:
                    error("%s Ornament Pad: values must be -100...100, not '%s'." % (self.name, v))
                o['pad'].append(v / 100)
            if len(o['pad']) == 1:
                o['pad'] += o['pad']
            
        elif cmd == 'PLACE':
            if opt not in ('ABOVE', 'BELOW', 'RANDOM'):
                error("%s Ornament Place: must be ABOVE, BELOW or RANDOM, not '%s'." %
                      (self.name, opt))
            o['place'] = opt

        elif cmd == 'VOLUME':
            t = stof(opt)
            if t > 0 and t <= 1000:
                o['volume'] = t / 100
            else:
                error("%s Ornament Volume: must be greater "
                      "than 0 and less that 1000, not '%s'." % (self.name, opt))

        elif cmd == 'BEATS':  
            o['beats'] = []
            if opt != 'ALL':
                for t in opt.split(','):
                    z = self.name  # this is for error msg in setBarOffset()
                    self.name = "%s Ornament beats" % z
                    o['beats'].append(self.setBarOffset(t))
                    self.name = z

        elif cmd == 'RSKIP':
            v = stof(opt)
            if v < 0 or v > 100:
                error("%s Ornament Rskip: must be 0 .. 100, not '%s'." % (self.name, v))
            o['rskip'] = v / 100

        elif cmd == 'RVOLUME':
            n1, n2 = mma.MMA.pat.getRndPair(opt, '%s Ornament' % self.name, -100, 100)
            o['rvolume'] = (n1 / 100., n2 / 100.)
            
        elif cmd == 'BARS':
            o['bars'] = []
            if opt != 'ALL':
                for t in opt.split(','):
                    v = stoi(t)
                    if v < 1 or v > gbl.seqSize:
                        warning("%s Ornament Bars: setting of %s may be "
                                "ignored since SeqSize is %s." %
                                (self.name, v, gbl.seqSize))
                    o['bars'].append(v - 1)

        elif cmd == 'OFFSET':
            v = stof(opt)
            if v < -gbl.BperQ or v > gbl.BperQ:
                error("%s Ornament Offset: must be %s .. %s, "
                      "not '%s' (note: value is MIDI ticks)." %
                      (self.name, -gbl.BperQ, gbl.BperQ, v))
            o['offset'] = v

        else:
            error("%s Ornament: '%s' is an unknown option." % (self.name, cmd))

    if mma.MMA.debug.debug:
        dPrint("%s Ornament: %s" % (self.name, getOrnOpts(self)))


def getNote(o, orig, scale, off):
    """ Calc a new note (the ornament) from the original. """

    if o['place'] == 'BELOW':
        off *= -1

    elif o['place'] == 'RANDOM':
        off = random.choice((1, -1))

    # Chromatic is easy.
    if o['chromatic']:
        return orig + off

    inoct = orig   # a copy of the original note adjusted to be in a 0..12
    oadjust = 0    # range. this lets us find it in the scale list.

    # Adjust 'inoct' to be in the scale list. The original note could
    # be '44' which is not in any scale list. So we need to adjust it
    # down by '12's until it's in range. Keep track of the number of times
    # we do this by also ajdusting 'oadjust' by the opposite value.
    while inoct > scale[-1]:
        oadjust += 12
        inoct -= 12
    while inoct < scale[0]:
        oadjust -= 12
        inoct += 12

    if inoct in scale:
        i = scale.index(inoct)    # this is the position of the note in the scale

        z = nn = i + off          # new note is just offset + old

        # adjust new into scale list
        while nn >= len(scale):
            oadjust += 12
            nn -= len(scale)
        while nn < 0:
            oadjust -= 12
            nn += len(scale)
        nn = scale[nn] + oadjust   # new note, plus octave

    else:
        # opps, not in scale, do chromatic adjustment
        # One place this can happen is with bass patterns using #/b options
        nn = orig + off

    return nn

def getOvolume(o, nvol):
    """ Calculate the note volume. This applies the ornament VOLUME and RVOLUME. """

    vol = int(nvol * o['volume'])
    r1 = int(vol * o['rvolume'][0])
    r2 = int(vol * o['rvolume'][1])
    if r1 or r2:
        if r2 < r1:
            r2, r1 = r1, r2
        vol += random.randrange(r1, r2)
    return vol

def getNOdur(o, noteDur, ornamentDur):
    """ Return an actual duration for the note and ornament. """

    return int(noteDur + (noteDur * o['pad'][0])), \
        int(ornamentDur + (ornamentDur * o['pad'][1]))


def doOrnament(self, nlist, scale, p):
    """ We pass the class structure and the [note duration] to ornament.
        Returns offset of the main note (for chords/harmony to sound right).
    """

    o = self.ornaments
    ornament = o['type']
    orn_duration = o['duration']
    orn_beats = o['beats']
    orn_rskip = o['rskip']
    orn_bars = o['bars']
    orn_offset = o['offset']
    orn_count = o['count']  # only effects gliss

    note, nvol = nlist[0]   # unpack the first note/duration passed.

    # if there is a beat list and the current note is not in the list
    # or if not the right seq bar, we send that note out without ornamentation.

    # skip over bars, beats and random
    if (orn_beats and p.offset not in orn_beats) \
            or (orn_bars and gbl.seqCount not in orn_bars) \
            or (orn_rskip and random.random() < orn_rskip):
        self.sendChord([[note, nvol]], p.duration, p.offset)
        retoffset = p.offset

    # start of the different option code.

    elif ornament == 'GRACE':  # short note before main
        gnote = getNote(o, note, scale, 1)
        odur = int(p.duration * orn_duration)
        ndur = p.duration - odur
        noteDur, ornamentDur = getNOdur(o, ndur, odur)
        ovol = getOvolume(o, nvol)
        self.sendChord([[gnote, ovol]], ornamentDur, p.offset + orn_offset)
        self.sendChord([[note, nvol]], noteDur, p.offset + odur)
        retoffset = p.offset + odur

    elif ornament == 'FALL':  # short note after main (reverse grace)
        gnote = getNote(o, note, scale, 1)
        odur = int(p.duration * orn_duration)
        ndur = p.duration - odur
        ovol =  getOvolume(o, nvol)
        noteDur, ornamentDur = getNOdur(o, ndur, odur)

        self.sendChord([[note, nvol]], ornamentDur, p.offset)
        self.sendChord([[gnote, ovol]], noteDur, p.offset + ndur + orn_offset)
        retoffset = p.offset

    elif ornament == 'TRILL':
        tnote = getNote(o, note, scale, 1)
        odur = int(p.duration * orn_duration)
        noteDur, _ = getNOdur(o, odur, odur)
        offset = p.offset
        count = range(int((p.duration / odur) / 2))  # number of pairs
        if not count:   # ensure at least 1 pair
            count = [1]
        for t in count:
            self.sendChord([[note, getOvolume(o, nvol)]], noteDur, offset)
            self.sendChord([[tnote, getOvolume(o, nvol)]], noteDur, offset + odur + orn_offset)
            offset += (odur + odur)
        retoffset = p.offset

    elif ornament == 'MORDENT':  # note, above, note
        onote = getNote(o, note, scale, 1)
        odur = int((p.duration * orn_duration) / 2)
        ndur = p.duration - odur - odur
        noteDur, ornamentDur = getNOdur(o, ndur, odur)

        self.sendChord([[note, getOvolume(o, nvol)]], ornamentDur, p.offset + orn_offset)
        self.sendChord([[onote, getOvolume(o, nvol)]], ornamentDur, p.offset + odur + orn_offset)
        self.sendChord([[note, nvol]], noteDur, p.offset + odur + odur)
        retoffset = p.offset + odur + odur

    elif ornament == 'TURN':  # above, note, below, note
        note1 = getNote(o, note, scale, 1)
        note2 = getNote(o, note, scale, -1)
        odur = int((p.duration * orn_duration) / 3)
        ndur = p.duration - (odur * 3)
        noteDur, ornamentDur = getNOdur(o, ndur, odur)
        self.sendChord([[note1, getOvolume(o, nvol)]],
                       ornamentDur, p.offset + orn_offset)
        self.sendChord([[note, getOvolume(o, nvol)]],
                       ornamentDur, p.offset + odur + orn_offset)
        self.sendChord([[note2, getOvolume(o, nvol)]],
                       ornamentDur, p.offset + (odur * 2) + orn_offset)
        self.sendChord([[note, nvol]], noteDur, p.offset + (odur * 3))
        retoffset = p.offset + (odur * 3)

    elif ornament == '3AFTER':     # note, 2above, 1above, 2above
        note1 = getNote(o, note, scale, 1)
        note2 = getNote(o, note, scale, 2)

        odur = int((p.duration * orn_duration) / 3)
        ndur = p.duration - (odur * 3)
        noteDur, ornamentDur = getNOdur(o, ndur, odur)

        self.sendChord([[note, nvol]], noteDur, p.offset)
        self.sendChord([(note2, getOvolume(o, nvol))],
                       ornamentDur, p.offset + ndur + orn_offset)
        self.sendChord([(note1, getOvolume(o, nvol))],
                       ornamentDur, p.offset + ndur + odur + orn_offset)
        self.sendChord([(note2, getOvolume(o, nvol))],
                       ornamentDur, p.offset + ndur + odur + odur + orn_offset)
        retoffset = p.offset 

    elif ornament == 'GLISS' or ornament == 'GLISSAFTER':  #  notes before or after
        
        # Make sure that each note in the gliss sounds for at least a tick.
        # Otherwise, it's really a bit of a waste. We need this since we don't
        # range test the count (except to make sure it's a positive integer).

        if  orn_duration * p.duration <  orn_count:
            orn_count = int(orn_duration * p.duration)

        # Create a list of gliss notes
        nlist = []
        for n in range(1, orn_count + 1):
            nn = getNote(o, note, scale, n)
            nlist.insert(0, nn)

        odur = int((p.duration * orn_duration) / orn_count)
        ndur = p.duration - (odur * orn_count)
        noteDur, ornamentDur = getNOdur(o, ndur, odur)
   
        y = p.offset

        if ornament == 'GLISSAFTER':
            self.sendChord([[note, nvol]], noteDur, y)
            retoffset = y
            y += ndur

        for n in nlist:
            self.sendChord([[n, getOvolume(o, nvol)]], ornamentDur, y )
            y += odur
        
        if ornament == 'GLISS':
            self.sendChord([[note, nvol]], noteDur, y)
            retoffset = y

    # when done we return the offset of the main note. This is only
    # used by CHORD so that the other notes in the chord line up.
    # Without this return the other notes in the chord cover up
    # the ornament.
    return retoffset



# rangeify.py

"""
The program "MMA - Musical Midi Accompaniment" and the associated
modules distributed with it are protected by copyright.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

The following code is borrowed from
       https://stackoverflow.com/questions/3429510

It has been modified to account for the MMA special
values of '?' indicating un-numbered lines and the
conversion of string values to integers.

"""
from itertools import count, groupby

def rangeify(l):
    for i,v in enumerate(l):
        try:
            l[i]=int(v)
        except ValueError:
            l[i]=-987654321

    G=(list(x) for _,x in groupby(l, lambda x,c=count(): next(c)-x))
    ret = ", ".join("-".join(map(str,(g[0],g[-1])[:len(g)])) for g in G)
    ret = ret.replace('-987654321', '?')
    return ret

#print( rangeify(['1', '2', '?', '4', '?', '?', '4','5']))




# docs.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import os
import time

import mma.MMA.midiC
import mma.MMA.grooves

from . import gbl
from   mma.MMA.common import *


def docDrumNames(order):
    """ Print LaTex table of drum names. """

    notenames = ['E\\flat', 'E', 'F', 'G\\flat', 'G', 'A\\flat',
                 'A', 'B\\flat', 'B', 'C', 'D\\flat', 'D'] * 5

    n = zip(mma.MMA.midiC.drumNames, range(27, len(mma.MMA.midiC.drumNames)+27), notenames)

    if order == "a":
        for a, v, m in sorted(n):
            print("\\insline{%s} {%s$^{%s}$}" % (a, v, m))

    else:
        for a, v, m in n:
            print("\\insline{%s} {%s$^{%s}$}" % (v, a, m))


def docCtrlNames(order):
    """ Print LaTex table of MIDI controller names. """

    n = zip(mma.MMA.midiC.ctrlNames, range(len(mma.MMA.midiC.ctrlNames)))

    if order == "a":
        for a, v in sorted(n):
            print("\\insline{%s} {%02x}" % (a, v))

    else:
        for a, v in n:
            print("\\insline{%02x} {%s}" % (v, a))


def docInstNames(order):
    """ Print LaTex table of instrument names. """

    n = zip(mma.MMA.midiC.voiceNames, range(len(mma.MMA.midiC.voiceNames)))
    if order == "a":
        for a, v in sorted(n):
            a = a.replace('&', '\&')
            print("\\insline{%s} {%s}" % (a, v))

    else:
        for a, v in n:
            a = a.replace('&', '\&')
            print("\\insline{%s} {%s}" % (v, a))


""" Whenever MMA encounters a DOC command, or if it defines
    a groove with DEFGROOVE it calls the docAdd() function.

    The saved docs are printed to stdout with the docDump() command.
    This is called whenever parse() encounters an EOF.

    Both routines are ignored if the -Dx command line option has
    not been set.

    Storage is done is in the following arrays.
"""

fname     = ''
author    = ""
notes     = ""
defs      = []
variables = []


def docAuthor(ln):
    global author

    author = ' '.join(ln)


def docNote(ln):
    """ Add a doc line. """

    global fname, notes

    if not gbl.createDocs or not ln:
        return

    # Grab the arg and data, save it

    fname = os.path.basename(gbl.inpath.fname)
    if notes:
        notes += ' '
    notes += ' '.join(ln)


def docVars(ln):
    """ Add a VARIABLE line (docs vars used in lib file)."""

    global fname, variables

    if not gbl.createDocs or not ln:
        return

    fname = os.path.basename(gbl.inpath.fname)
    variables.append([ln[0], ' '.join(ln[1:])])


def docDefine(ln):
    """ Save a DEFGROOVE comment string.

        Entries are stored as a list. Each item in the list is
        complete groove def looking like:
        defs[ [ Name, Seqsize, Description, [ [TRACK,INST, [Sequences...] ]...]] ...]

    """

    global defs

    # Skip if not creating docs
    if not gbl.createDocs:
        return

    des = ' '.join(ln[1:])  
    if des[-1] != '.':  # Add a '.' to the end of all descriptions
        des += '.'
    l = [ln[0], gbl.seqSize, des]
    for a in sorted(gbl.tnames.keys()):
        c = gbl.tnames[a]
        if c.sequence and len(c.sequence) != c.sequence.count(None):
            if c.vtype == 'DRUM':
                v = [mma.MMA.midiC.valueToDrum(x) for x in c.toneList]
            else:
                v = [mma.MMA.midiC.valueToInst(x) for x in c.voice]
            seq = [c.formatPattern(c.sequence[x]) for x in range(gbl.seqSize)]
            l.append([c.name, v, seq])

    defs.append(l)


def docDump():
    """ Print the LaTex docs. """

    global fname, author, notes, defs, variables

    if gbl.createDocs == 1:    # latex docs
        if notes:
            notes = notes.replace("<P>", "\\\\[.5ex]")
            notes = notes.replace("<p>", "\\\\[.5ex]")
            if fname.endswith(gbl.EXT):
                fname = '.'.join(fname.split('.')[:-1])
            print("\\filehead{%s}{%s}\n" % (totex(fname), totex(notes)))

        if variables:
            print("  \\variables{")
            for l in variables:
                print("     \\insvar{%s}{%s}" % (totex(l[0]), totex(l[1])))
            print("  }\n")

        if defs:
            for l in defs:
                alias = mma.MMA.grooves.getAlias(l[0])
                if alias:
                    if len(alias) > 1:
                        alias = "Aliases: %s" % alias
                    else:
                        alias = "Alias: %s" % alias
                else:
                    alias = ''
                print("     \\instable{%s}{%s}{%s}{%s}{" % 
                    (totex(l[0]), totex(l[2]), l[1], alias))
                for c, v, s in l[3:]:  # we ignore the seqence data here
                    print("       \\insline{%s}{%s}" % (c.title(), totex(v[0])))
                print("     }")

    elif gbl.createDocs == 2:    # html docs
        if notes:
            print('<!-- Auto-Generated by MMA on: %s -->' % time.ctime())
            print('<HTML>')
            print('<BODY  BGCOLOR="#B7DFFF" Text=Black>')
            if fname.endswith(gbl.EXT):
                fname = '.'.join(fname.split('.')[:-1])
            print("<H1>%s</H1>" % fname.title())
            print("<P>%s" % notes)

        if variables:
            print("<P>")
            print('<Table Border=3 CELLSPACING=0 CELLPADDING=5 BGColor="#eeeeee" Width="60%">')
            print('  <TR><TD>')
            print('    <H2> Variables </H2> ')
            print('  </TD></TR>')
            print('  <TR><TD>')
            print('    <Table CELLSPACING=0 CELLPADDING=5 BGColor="#eeeeee" Width="100%">')
            for l in variables:
                print("       <TR>")
                print("          <TD Valign=Top> <B> %s </B> </TD> " % l[0])
                print("          <TD Valign=Top> %s </TD>" % l[1])
                print("       </TR>")
            print('    </Table>')
            print('  </TD></TR>')
            print('</Table>')

        if defs:
            print("<ul>")
            for l in defs:
                print("<LI><A Href=#%s>%s</a>" % (l[0], l[0]))
            print("</ul>")
            for l in defs:
                gg = l[0]
                iname = os.path.basename(gbl.infile)
                iname, ext = os.path.splitext(iname)
                gfile = "%s_%s.html" % (iname, gg.lower())
                print('<!-- GROOVE=%s FILE=%s SRC=%s -->' % (gg.lower(), gfile, gbl.infile))
                print('<A Name=%s></a>' % gg)
                print('<Table Border=3 CELLSPACING=0 CELLPADDING=5 BGColor="#eeeeee" Width="60%">')
                print('  <TR><TD>')
                print('    <H2> <A Href=%s> %s </a> </H2> ' % (gfile, l[0]))
                alias = mma.MMA.grooves.getAlias(l[0])
                if alias:
                    if len(alias) > 1:
                        ll = "Aliases"
                    else:
                        ll = "Alias"
                    print(' <H4> %s: %s </H4>' % (ll, alias))

                print('    %s <B>(%s)</B> ' % (l[2], l[1]))
                print('  </TD></TR>')
                print('  <TR><TD>')
                print('    <Table CELLSPACING=0 CELLPADDING=5 BGColor="#eeeeee" Width="10%">')
                for c,v,s in l[3:]:
                    print("       <TR><TD> %s </TD> <TD> %s </TD></TR>" % (c.title(), v[0]))
                print('    </Table>')
                print('  </TD></TR>')
                print('</Table>')
            print('\n</Body></HTML>')

    elif gbl.createDocs == 3:    # sequence table
        for l in defs:
            print("GROOVE %s" % l[0])
            print("DESCRIPTION %s" % l[2])
            print("SIZE %s" % l[1])
            for c, v, s in l[3:]:
                print("TRACK %s" % c)
                print("VOICE %s" % v)
                print("SEQ %s" % s)

    elif gbl.createDocs == 99:  # creating entry for groove browser
        for a, b in (("``", '"'), ("''", '"'), ('  ', ' ')):
            notes = notes.replace(a, b)
        if not notes:
            notes = "No header available ... please add DOC to file"
        print(notes)

        for l in defs:
            print("%s\n %s" % (l[0], l[2]))

    else:
        return

    defs = []
    variables = []
    notes = ""
    author = ""


def totex(s):
    """ Parse a string and quote tex stuff.

        Also handles proper quotation style.
    """

    s = s.replace("$", "\$")
    s = s.replace("*", "$*$")
    s = s.replace("_", "\\_")
    #s = s.replace("\\", "\\\\")
    s = s.replace("#", "\\#")
    s = s.replace("&", "\\&")
    q = "``"
    while s.count('"'):
        s = s.replace('"', q, 1)
        if q == "``":
            q = "''"
        else:
            q = "``"

    return s


def htmlGraph(f):
    """ Print (stdout) an html file representing a graph and details of the current groove."""

    global fname, author, notes, variables, defs

    def getAbsPair(x1, x2):
        if abs(x1) == abs(x2):
            return "%s" % abs(x1)
        else:
            return "%s,%s" % (x1, x2)

    def docol(lab, data):
        if lab == '':
            return
        
        if not isinstance(data, list):
            data = [data]
        print("  <td width=50%> ")
        print("%s: " % lab)
        if len(set(data)) == 1:
            print(str(data[0]))
        else:
            print('&nbsp;&nbsp; '.join([str(x) for x in data]))
        print("  </td>")

    def dorow(a, a1, b, b1):
        print("<tr>")
        docol(a, a1)
        docol(b, b1)
        print("</tr>")

    defs = []
    variables = []
    notes = ""
    author = ""
    desc = ""

    if '/' in f:
        u, f = f.rsplit('/', 1)
        mma.MMA.parse.usefile([u])
    mma.MMA.grooves.groove([f])

    groove = mma.MMA.grooves.currentGroove.title()

    for a in defs:
        if a[0].upper() == groove.upper():
            desc = a[2]

    print('<!-- Auto-Generated by MMA on: %s -->' % time.ctime())
    print('<HTML>')
    print('<BODY  BGCOLOR="#B7DFFF" Text=Black>')

    #if fname.endswith(gbl.EXT):
    #    fname='.'.join(fname.split('.')[:-1])

    print("<h2>File: %s</h2>" % fname)
    print("<h2>Groove: %s</h2>" % groove)
    print("<p><b>Notes:</b> %s" % notes)
    print("<p><b>Author:</b> %s" % author)
    print("<p><b>Description:</b> %s" % desc)

    print("<p><Table Border=0 Width=75%>")
    dorow("SeqSize", gbl.seqSize, "Time (beats per bar)", gbl.QperBar)
    print("</Table>")

    if variables:
        print("<p> <b>Variables</b>")
        print('<Table Border=.2em Cellspacing=5 Cellpadding=5  Width="90%">')
        for l in variables:
            print("          <TD Valign=Top> <B> %s </B> </TD> " % l[0])
            print("          <TD Valign=Top> %s </TD>" % l[1])
            print("       </TR>")
        print('</Table>')

    for t in sorted(gbl.tnames.keys()):
        trk = gbl.tnames[t]
        if not trk.sequence or len(trk.sequence) == trk.sequence.count(None):
            continue

        print("<p> <b>Track Name: %s</b>" % t.title())
        print("<p><Table Border=0 Width=75%>")
        if trk.vtype == "DRUM":
            v = [mma.MMA.midiC.valueToDrum(x) for x in trk.toneList]
        else:
            v = [mma.MMA.midiC.valueToInst(x) for x in trk.voice]
        dorow("Voice/Tones", v, "Articulate", trk.artic)
        v = [str(int(x * 100)) for x in trk.volume]
        if trk.vtype != "DRUM":
            oct = [x // 12 for x in trk.octave]
            dorow('Unify', trk.unify, "Octave", oct)
        dorow("Volume", v, "Harmony", trk.harmony)
        v1 = [str(int(x * 100)) for x in trk.rSkip]
        v2 = [getAbsPair(int(x1 * 100), int(x2 * 100)) for x1, x2 in trk.rVolume]
        dorow("Rskip", v1, "Rvolume", v2)
        v1 = [getAbsPair(x1, x2) for x1, x2 in trk.rTime]
        if trk.seqRnd:
            v2 = "On"
        else:
            v2 = "Off"
        dorow("Rtime", v1, "SeqRND", v2)
        if trk.vtype == 'CHORD':
            vv = trk.voicing.mode
            v = 'Voicing'
        else:
            vv = v = ''
        strm = []
        for z in trk.strum:
            if z is None:
                strm.append( "None")
            else:
                strm.append( getAbsPair(z[0], z[1]))
        dorow("Strum", strm, v, vv)
        print("</Table>")
 
        pointx = 2.5
        pointPerS = pointx * gbl.QperBar
        pointy = 5
        boxx = gbl.seqSize * pointPerS
        boxy = pointy

        print(r'<div style="position:relative;background-color:#99bdf4;')
        print(r'padding:0; border:.1em solid black; left:5em;')
        print(r'height:%sem; width:%sem">' % (boxy, boxx))

        if gbl.seqSize > 1:
            for a in range(1, gbl.seqSize):
                print(r'<img style="position:absolute; bottom:0; left:%sem;' % (a * pointPerS))
                print(r'width:.05em; height:%sem; border:.05em solid white"' % pointy)
                print(r'src="../black.gif">')

        for a in range(gbl.seqSize):
            s = trk.sequence[a]
            if not s:
                continue
            for p in s:
                bwidth = p.duration * pointx * (trk.artic[a] / 100.) / gbl.BperQ 
                if bwidth < .1:
                    bwidth = .1
                offset = (p.offset * pointx) // gbl.BperQ + (a * pointPerS)
                if trk.vtype == 'CHORD' or trk.vtype == 'PLECTRUM':
                    # A plectrum track with all 0 volumes is a mute, so ignore.
                    if trk.vtype == 'PLECTRUM' and sum(p.vol) == 0:
                        continue
                    ll = len(p.vol) - p.vol.count(0)
                    vol = sum(p.vol) // ll
                else:
                    vol = p.vol
                height = (vol * pointy) // 127

                print(r'<img style="position:absolute; border:.02em solid red;bottom:0; ')
                print(r'left:%sem; width:%sem; height:%sem"'  % ( offset, bwidth, height))
                print(r' src="../blue.gif">')

        print(r'</div>')

    print('</Body></HTML>')

    sys.exit(0)




# auto.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""


import os
import sys
import pickle

import mma.MMA.midi
import mma.MMA.parse
import mma.MMA.grooves
import mma.MMA.swing

import mma.MMA.paths

from . import gbl
from mma.MMA.common import *

grooveDB = []        # when filled in it becomes [['dir', dict-db], ..]
mmadir = ".mmaDB"   # constant, name of the lib database file
mkGrooveList = []    # a list of grooves defined in current file


def updateGrooveList(n):
    """ Called from parse when new grooves are defined in a -g. """

    global mkGrooveList

    mkGrooveList.append(n)


def libUpdate():
    """ Update the mma library database file(s) with -g or -G option.

        This is called from the main program after the initialization
        and other option parsing. No RETURN.
    """

    global grooveDB, mkGrooveList

    processedFiles = []
    dupMessage = []
    grooveCount = 0
    fileCount = 0
    noAccess = 0

    print( "Creating MMA groove directory database(s). Standby..." )

    """ mma.MMA.paths.libDirs contains a complete list of all the directories in
        the libPath; we create a .mmaDB file for each one.
    """

    if not mma.MMA.paths.libDirs:
        mma.MMA.paths.expandLib()
    for lib in mma.MMA.paths.libDirs:

        gdDate = None
        grooveDB = [[lib, {}]]

        # load up our database with this directory's DB file, skip if -G

        if gbl.makeGrvDefs == 1:
            g = loadDB(lib)
            if g:
                grooveDB = [[lib, g]]
                gdDate = os.path.getmtime(os.path.join(lib, mmadir))

        # -G
        # If the database is really screwed up it's possible to
        # have old links not deleted. Easy way is to completely
        # delete each .mmaDB file as we progress. If we can't delete
        # it,it's probably due to a permission error. So, just dump out.

        elif gbl.makeGrvDefs == 2:
            file = os.path.join(lib, mmadir)
            if os.path.exists(file):
                try:
                    os.remove(file)
                except:
                    error("Unable to delete existing database file '%s'. Are you root?" \
                            % file)

        db = grooveDB[0][1]

        print("     Processing library directory '%s'." % lib)

        """ Get a list of the files in this directory. If the list
            includes a file called 'MMAIGNORE' the entire directory
            is ignored. Otherwise, each file in the
            directory ending in 'mma' is parsed for groove defs.
        """

        dirfiles = os.listdir(lib)
    
        if "MMAIGNORE" in dirfiles:
            print("Skipping: %s" % p)
            return

        for fn in sorted(dirfiles):  # sort is just to pretty up console display

            # Ignore hidden files, emacs auto-save and non-mma files

            if fn.startswith('.') or fn.startswith('#') or not fn.endswith(gbl.EXT):
                continue

            f = os.path.join(lib, fn)      # Create full path name
            processedFiles.append(f)     # a list of ALL processed (complete paths)

            if gdDate and f in db and os.path.getmtime(f) < gdDate:
                print("        Skipped: %s" % f )
                grooveCount += len(db[f])
                continue

            if f in db:
                print("       Updating: %s" % f)
            else:
                print("       Creating: %s" % f)

            mkGrooveList = []
            mma.MMA.grooves.grooveClear([])
            gbl.mtrks = {}
            mma.MMA.swing.mode = 0
            for c in gbl.midiAssigns.keys():
                gbl.midiAssigns[c] = []
            for a, v in enumerate(gbl.midiAvail):
                gbl.midiAvail[a] = 0
            gbl.mtrks[0] = mma.MMA.midi.Mtrk(0)

            gbl.tnames = {}

            mma.MMA.parse.parseFile(f)    # read current file, grab grooves

            fileCount += 1            # just so we can report to user
            grooveCount += len(mkGrooveList)
            db[f] = mkGrooveList

        # Strip out defs of deleted (not found) files. Only on -g (update mode)

        if gbl.makeGrvDefs == 1:
            for f in db.keys():
                if f not in processedFiles:
                    print("       Deleting: %s" % f)
                    del db[f]

        try:
            outpath = open(os.path.join(lib, mmadir), 'wb')
        except IOError:
            print("     Skipping '%s', no write access." % lib)
            noAccess += 1
            continue

        outpath.write(b"### mmaDB ... AUTOGENERATED BINARY DATA. "
                      b"DO NOT EDIT!!!\n")
        pickle.dump(db, outpath, protocol=2)  # proto2 forces compat between py2 and py3
        outpath.close()
        
        # check the database we just saved for duplicate entries.

        dbKeys = sorted(db.keys())
        idx = 1
        for k in dbKeys[:-1]:
            for kk in dbKeys[idx:]:
                for g in db[k]:
                    if g in db[kk]:
                        dupMessage.append("   Lib %s: %s & %s\n" % \
                                              (lib, os.path.basename(k),
                                               os.path.basename(kk)))
                        break
            idx += 1

    print("\nDatabase update complete."
            "\n     Files processed: %s" 
            "\n     Total number of grooves: %s\n" % (fileCount, grooveCount ))
    
    if noAccess:
        error ("You probably need to be 'root' to properly update your libraries.")
    
    if dupMessage:
        msg = ["Warning: Duplicate groove definitions found in:\n"]
        for a in dupMessage:
            msg.append(a)
        print(' '.join(msg))
    
    sys.exit(0)


def loadDB(d):
    """ Read a database file into memory.

        We're assuming that not much goes wrong here...if we don't find
        the database we return a Null.
    """

    try:
        infile = os.path.join(d, mmadir)
        f = open(infile, "rb")
        f.readline()    # Read/discard comment line
        g = pickle.load(f)
        f.close()
    except IOError:
        g = {}
    except ValueError:
        error("Incompatible database found. Maybe a result"
              "  of using different versions of python to"
              "  create/read, disk corruption or someone"
              "  playing games with the DB2 files."
              "  Recreate database by using 'mma -G' as root.")

    # If we can't find the directory, advise user to recreate the DB
    # Note, not an error ... should it be? Note, the compile will probably
    # end soon anyway since mma won't be able to find the requested groove.

    dirs = [os.path.dirname(i) for i in g.keys()]

    for a in set(dirs):
        if not os.path.exists(a):
            warning("There appears to be an error in the database.\n"
                    "  Directory '%s' cannot be found.\n"
                    "  Please try to execute the command 'mma -G' as root." % a)

    return g


#################################################################

def searchDB(targ, skip=None):
    """ Scan the database for the desired groove name. """
    
    # Note the way we jump out the doubly nested loop.
    # Yes, this is pythonic and fast.
    try:
        for dir, g in grooveDB:
            for filename, namelist in g.items():
                if targ in namelist:
                    raise StopIteration

    except StopIteration:
        return os.path.join(dir, filename), targ.upper()

    return ('', targ)
    
def findGroove(targ):
    """ Try to auto-load a groove from the library.

        The groove name (targ) can be a simple name (ie. swingsus) or
        an entended name which includes a file and directory
         (ie. swing:swingsus or stdlib/swing:swingsus).
    """

    global grooveDB
    ret = (None, targ.upper())   # not found return

    # If no existing DB we load them from expanded libPath

    if not grooveDB:
        grooveDB = []
        if not mma.MMA.paths.libDirs:
            mma.MMA.paths.expandLib()
        for d in mma.MMA.paths.libDirs:
            g = loadDB(d)
            if g:
                grooveDB.append([d, g])

        if not grooveDB:   # BS value so we don't keep trying to load
            grooveDB = [['', {}]]

    # Check for extended notation. This is [DIR/DIR][FILE]:GROOVE

    # Split the filename from the groove name.
    # The complete spec for a groove name is dir/subdir/file:groove

    if ':' in targ:
        dirfile, targ = targ.split(':', 1)
        dirpath = mma.MMA.paths.findLibFile(dirfile)
        targ = targ.upper()
        if dirpath == None:
            error("The file '%s' does not exist in libraries." % dirfile)
            
        # user is by-passing the default libs ... just pass the file/groovename back
        if dirfile.startswith('.') or os.path.isabs(dirfile):
            ret = (dirfile, targ)
            
        # did the user shortcut the extended format by using DIR:groove (without a
        # filename. That's fine, search the database for the specifed dir.
        # Only problem is that if you have duplicate grooves defined in the
        # lib the first one found will be used.
        elif os.path.isdir(dirpath):
            ret = searchDB(targ, dirfile)
                            
        else:
            # Does the library file exist at all? Check all the directories
            # set up from the expanded libPath. An example search
            # might be for "stdlib/swing:swingsus". The paths will
            # be:   1. stdlib  (no stdlib/swing.mma here)
            #      ..  lib   YES... found stdlib/swing.mma here

            fpath = mma.MMA.paths.findLibFile(dirfile)

            if not fpath:
                error("Can't locate library file: %s! Groove: %s can't be found either."
                      % (dirfile, targ))

            if fpath.endswith(gbl.EXT) and not dirfile.endswith(gbl.EXT):
                dirfile += gbl.EXT

            ret = (dirfile, targ)

    else:
        # Just lookup a normal groove load from a file. We scan
        # the database and search for the groove name (targ).


        targ = targ.upper()
        ret = searchDB(targ)
   
    return ret



# after.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

After storage, setting and command insertion.

"""

from . import gbl
from mma.MMA.common import *
import mma.MMA.debug

class AfterData:
    def __init__(self):
        self.bar = None
        self.cmd = None
        self.repeat = None
        self.count = None
        self.lineno = -1
        self.id = None

afterData = []

def create(ln):
    """ Set an After event. """

    global afterData

    dat = AfterData()
    selected = []

    ln, opts = opt2pair(ln, toupper=False, notoptstop=True)

    for cmd, opt in opts:
        cmd = cmd.upper()
        if cmd == 'BAR':
            if opt.upper() == 'EOF':
                dat.bar = 'EOF'
            else:
                opt = stoi(opt)
                if opt < 1:
                    error("After: Destination bar must be positive, not '%s'." % opt)
                    dat.bar = opt
            selected.append('Bar')
                
        elif cmd == 'REPEAT':
            opt = stoi(opt)
            if opt < 1:
                error("After: Repeat value must be positive, not '%s'." % opt)
            dat.repeat = opt
            dat.bar = opt + gbl.barNum 
            selected.append('Repeat')

        elif cmd == 'COUNT':
            opt = stoi(opt)
            if opt < 1:
                error("After: Count must be positive, not '%s'." % opt)
            dat.count = opt
            dat.bar = opt + gbl.barNum 
            selected.append('Count')

        elif cmd == 'ID':   # set a string id for this event. 
            dat.id = opt.upper()

        elif cmd == 'REMOVE':  # delete all saved items with id==opt.
            t=len(afterData)
            afterData = [ x for x in afterData if x.id != opt.upper() ]
            if len(opts) > 1 or ln:
                warning("After: Remove is ignoring other options/cmds in line.")
            if len(afterData) == t:
                warning("After: No events with ID=%s found to delete." % opt)
            return   # ignore any other commands

        else:
            error("After: Unknown option '%s'." % cmd)


    if not ln:
        error("After: No command given. It's needed.")

    if len(selected) == 0:
        error("After: need one of 'Count', 'Bar', or 'Repeat'.")

    if len(selected) > 1:
        error("After: Options conflict; can't combine %s." % ' & '.join(selected))

    dat.cmd  = ln
    dat.lineno = gbl.lineno

    if dat.bar == 'EOF':
         gbl.inpath.pushEOFline(dat.cmd)
    else:
        afterData.append(dat)

    if mma.MMA.debug.debug:
        dPrint("After: Added event '%s' at bar %s." % (' '.join(dat.cmd), dat.bar))

def check():
    """ Before reading any input, we check to see if any AFTER events have been
        created and if we need to process them now. 
    """

    global afterData
    stuff = []
    elns = []

    nn = gbl.barNum
    for dat in afterData:
        if dat.bar == nn:
            stuff.append(dat.cmd)
            elns.append(dat.lineno)
            if dat.repeat:
                dat.bar = gbl.barNum + dat.repeat

    if stuff:
        # Push our AFTER command to input.
        gbl.inpath.push(stuff, elns)

        # delete any discarded AFTER events
        afterData = [ x for x in afterData if x.bar != nn]




# sync.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

Functions/variables for the sync stuff.
  Sync START puts a midi on/off event at the start of every track,
  END pads all tracks out to the same length. This is a "be kind"
  process for certain synths and notation converters.
"""

from mma.MMA.common import *
import mma.MMA.debug

synctick    =  0      # flag, set if we want a tick on all tracks at offset 0
endsync     =  0      # flag, set if we want a eof sync

syncTone = [80, 90]  # tone/velocity for -0 option. Changable from setSyncTone

def synchronize(ln):
    """ Set synchronization in the MIDI. This is called by the SYNCRONIZE option
        and the command line -0 and -1 options.
    """

    global synctick, endsync
    
    if not ln:
        error("SYNCHRONIZE: requires args END and/or START.")

    for a in ln:
        if a.upper() == 'END':
            endsync = 1
        elif a.upper() == 'START':
            synctick = 1
        else:
            error("SYNCHRONIZE: expecting END or START")
    

def setSyncTone(ln):
    """ Parser routine, sets tone/velocity for the -0 sync tone. """

    global syncTone
    
    notopts, ln = opt2pair(ln)

    if notopts or not ln:
        error("SetSyncTone: Expecting option pairs: Tone=xx Velocity=xx Volume=xx.")

    for cmd, opt in ln:
        cmd = cmd.upper()

        if cmd == "TONE":
            t = stoi(opt)
            if t < 0 or t > 127:
                error("SetSyncTone: Tone must be 0..127, not %s." % t)
            syncTone[0] = t

        elif cmd in ("VELOCITY", "VOLUME"):
            t = stoi(opt)
            if t < 1 or t > 127:
                error("SetSyncTone: Velocity must be 1..127, not %s." % t)
            syncTone[1] = t

        else:
            error("SetSyncTone: Expecting options: Tone, Velocity, Volume. Not %s"
                  % cmd)

    if mma.MMA.debug.debug:
        dPrint("SetSyncTone: Tone=%s, Velocity=%s" % tuple(syncTone))




# patChord.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""


import random

import mma.MMA.notelen
import mma.MMA.ornament
from . import gbl
from mma.MMA.common import *
from mma.MMA.pat import PC, Pgroup
import mma.MMA.debug

import copy


class Voicing:
    def __init__(self):
        self.mode = None
        self.range = 12
        self.center = 4
        self.random = 0
        self.percent = 0
        self.bcount = 0
        self.dir = 0


class Chord(PC):
    """ Pattern class for a chord track. """

    vtype = 'CHORD'
    sortDirection = 0   # used for tracking direction of strummed chords

    def __init__(self, ln):
        self.voicing = Voicing()
        PC.__init__(self, ln)

    def saveGroove(self, gname):
        """ Save special/local variables for groove. """

        PC.saveGroove(self, gname)  # create storage. Do this 1st.
        self.grooves[gname]['VMODE'] = copy.deepcopy(self.voicing)

    def restoreGroove(self, gname):
        """ Restore special/local/variables for groove. """

        self.voicing = self.grooves[gname]['VMODE']
        PC.restoreGroove(self, gname)

    def clearSequence(self):
        """ Set some initial values. Called from init and clear seq. """

        PC.clearSequence(self)
        self.voicing = Voicing()
        # .direction was set in PC.clear.. we're changing to our default
        self.direction = seqBump(['UP'])

    def setVoicing(self, ln):
        """ set the Voicing Mode options.  Only valid for CHORDS. """

        notopt, ln = opt2pair(ln, toupper=True)

        if notopt:
            error("Voicing: Each Voicing option must be a OPT=VALUE pair.")

        for mode, val in ln:
            if mode == 'MODE':
                valid = ("-", "OPTIMAL", "NONE", "ROOT", "COMPRESSED",
                         "INVERT", "KEY", "KEY2", "ROOTKEY")

                if not val in valid:
                    error("Valid Voicing Modes are: %s" % " ".join(valid))

                if val in ('-', 'NONE', 'ROOT'):
                    val = None

                # If we're setting a voicing mode check to see if we've
                # previously set invert or compress ... warning
                if val and (any(self.invert) or any(self.compress)):
                    warning("Setting both VoicingMode and "
                            "Invert/Compress is not a good idea")
                        
                # When we set voicing mode we always reset this. This forces
                # the voicingmode code to restart its rotations.

                self.lastChord = []

                self.voicing.mode = val

            elif mode == 'RANGE':
                val = stoi(val, "VOICING RANGE %s: Arg must be a value" % self.name)

                if val < 1 or val > 30:
                    error("Voicing Range: Arg out-of-range; must be 1 to 30, not '%s'." % val)

                self.voicing.range = val

            elif mode == 'CENTER':

                val = stoi(val, "Argument for %s VOICING CENTER must be a value."
                           % self.name)

                if val < 1 or val > 12:
                    error("VOICING CENTER: arg out-of-range; must be 1 to 12, not '%s'." % val)

                self.voicing.center = val

            elif mode == 'RMOVE':
                val = stoi(val, "Argument for %s VOICING RANDOM must be a value" % self.name)

                if val < 0 or val > 100:
                    error("VOICING RANDOM: arg must be 0 to 100, not %s" % val)

                self.voicing.random = val
                self.voicing.bcount = 0

            elif mode == 'MOVE':
                val = stoi(val, "Argument for %s VOICING MOVE  must be a value" % self.name)

                if val < 0:
                    error("VOICING MOVE: bar count must >= 0, not %s" % val)
                if val > 20:
                    warning("VOICING MOVE: bar count '%s' is quite large" % val)

                self.voicing.bcount = val
                self.voicing.random = 0

            elif mode == 'DIR':
                val = stoi(val, "Argument for %s VOICING DIR must be a value" % self.name)

                if not val in (1, 0, -1):
                    error("VOICING MOVE: Dir must be -1, 0 or 1, not '%s'." % val)

                self.voicing.dir = val

            else:
                error("Voicing does not have a '%s' option" % mode)
                
        if mma.MMA.debug.debug:
            v = self.voicing
            dPrint("Set %s Voicing MODE=%s RANGE=%s CENTER=%s RMOVE=%s MOVE=%s DIR=%s " %
                  (self.name, v.mode, v.range, v.center, v.random, v.bcount, v.dir))

    def setDupRoot(self, ln):
        """ set/unset root duplication. Only for CHORDs """

        ln = lnExpand(ln, '%s DupRoot' % self.name)
        tmp = []

        for n in ln:
            ll = []
            for v in n.split(','):
                v = stoi(v, "%s DupRoot: Argument must be a value, not '%s'."
                            % (self.name, v))

                if v < -9 or v > 9:
                    error("%s DupRoot: '%s' out-of-range; must be -9 to 9." % (self.name, v))

                if v:
                    ll.append(v * 12)

            tmp.append(ll)

        self.dupRoot = seqBump(tmp)

        if mma.MMA.debug.debug:
            dPrint("%s DupRoot set to: %s" % (self.name, self.getDupRootSetting()))

    def getDupRootSetting(self):
        """ Need to convert nested list ints to string. """

        ret = ""
        for l in self.dupRoot:
            if not l:
                l = [0]
            ret += ','.join([str(x // 12) for x in l]) + "  "
        return ret.strip()

    def getPgroup(self, ev):
        """ Get group for chord pattern.

        Tuples: [start, length, volume (,volume ...) ]
        """

        if len(ev) < 3:
            error("There must be at least 3 items in each group "
                  "of a chord pattern definition, not <%s>" % ' '.join(ev))

        a = Pgroup()

        a.offset = self.setBarOffset(ev[0])
        a.duration = mma.MMA.notelen.getNoteLen(ev[1])

        vv = ev[2:]
        if len(vv) > 8:
            error("Only 8 volumes are permitted in Chord definition, not %s" % len(vv))

        a.vol = [0] * 8
        for i, v in enumerate(vv):
            v = stoi(v, "Expecting integer in volume list for Chord definition")
            a.vol[i] = v

        for i in range(i + 1, 8):  # force remaining volumes
            a.vol[i] = v

        return a

    def restart(self):
        self.ssvoice = -1
        self.lastChord = None

    def chordVoicing(self, chord, vMove):
        """ Mangle chord notes for voicing options. """

        sc = self.seq
        vmode = self.voicing.mode

        if vmode == "OPTIMAL":  # Optimal voicing algorithm by Alain Brenzikofer.

            # Initialize with a voicing around centerNote
            chord.center1(self.lastChord)

            # Adjust range and center
            if not (self.voicing.bcount or self.voicing.random):
                chord.center2(self.voicing.center, self.voicing.range // 2)

            # Move voicing
            elif self.lastChord:
                if (self.lastChord != chord.noteList) and vMove:
                    chord.center2(self.voicing.center, self.voicing.range // 2)
                    vMove = 0

                    # Update voicingCenter
                    sum = 0
                    for n in chord.noteList:
                        sum += n
                    c = sum // chord.noteListLen

                    # If using random voicing move it it's possible to
                    # get way off the selected octave. This check ensures
                    # that the centerpoint stays in a tight range.
                    # Note that if using voicingMove manually (not random)
                    # it is quite possible to move the chord centers to very
                    # low or high keyboard positions!

                    if self.voicing.random:
                        if c < -4:
                            c = 0
                        elif c > 4:
                            c = 4
                    self.voicing.center = c

        elif vmode == "KEY":
            chord.keycenter()

        elif vmode == "KEY2":
            chord.key2center()

        elif vmode == "ROOTKEY":
            chord.rootkey()

        elif vmode == "COMPRESSED":
            chord.compress()

        elif vmode == "INVERT":
            if chord.rootNote < -2:
                chord.invert(1)
            elif chord.rootNote > 2:
                chord.invert(-1)
            chord.compress()

        self.lastChord = chord.noteList[:]

        return vMove

    def trackBar(self, pattern, ctable):
        """ Do a chord bar. Called from self.bar() """

        sc = self.seq
        unify = self.unify[sc]

        # Set voicing move ONCE at the top of each bar.
        # The voicing code resets vmove to 0 the first
        # time it's used. That way only one movement is
        # done in a bar.

        vmove = 0

        if self.voicing.random:
            if random.randrange(100) <= self.voicing.random:
                vmove = random.choice((-1, 1))
        elif self.voicing.bcount and self.voicing.dir:
            vmove = self.voicing.dir


        for p in pattern:
            tb = self.getChordInPos(p.offset, ctable)

            if tb.chordZ:
                continue

            dupRoot = self.dupRoot[sc]

            vmode = self.voicing.mode
            vols = p.vol[0:tb.chord.noteListLen]

            # Limit the chord notes. This works even if THERE IS A VOICINGMODE!

            if self.chordLimit:
                tb.chord.limit(self.chordLimit)

            # Compress chord into single octave if 'compress' is set
            # We do it here, before octave, transpose and invert!
            # Ignored if we have a VOICINGMODE.

            if self.compress[sc] and not vmode:
                tb.chord.compress()

            # Do the voicing stuff.

            if vmode:
                vmove = self.chordVoicing(tb.chord, vmove)

            # Invert.

            if self.invert[sc]:
                tb.chord.invert(random.randrange(self.invert[sc][0], self.invert[sc][1]+1))

            # Voicing adjustment for 'jazz' or altered chords. If a chord (most
            # likely something like a M7 or flat-9 ends up with any 2 adjacent
            # notes separated by a single tone an unconfortable dissonance results.
            # This little check compares all notes in the chord and will cut the
            # volume of one note to reduce the disonance. Usually this will be
            # the root note volume being decreased.

            nl = tb.chord.noteList
            l = len(nl)
            for j in range(l - 1):
                r = nl[j]
                for i in range(j + 1, l):
                    if nl[i] in (r - 1, r + 1, r - 13, r + 13) and vols[i] >= vols[0]:
                        vols[j] = vols[i] // 2
                        break

            loo = list(zip(nl, vols))    # this is a note/volume array of tuples

            # DupRoot adds additional root tones in
            # different octaves to fatten chord sounds.

            if dupRoot:

                # Volume for the dups is adjusted by taking the
                # average of the non-zero volumes in the chord
                # and adjusting with the harmonyVolume.

                v = [x for x in vols if x > 0]
                if v:   # just in case we have a no-volume chord
                    v = int((sum(v) // len(v)) * self.harmonyVolume[sc])

                    root = tb.chord.rootNote  # true root of chord
                    for nn in dupRoot:
                        loo.append(((nn + root), v))

            # For strum we need to know the direction. Note that the direction
            # is saved for the next loop (needed for alternating in BOTH).

            sd = self.direction[sc]
            if sd == 'BOTH':
                if self.sortDirection:
                    self.sortDirection = 0
                else:
                    self.sortDirection = 1
            elif sd == 'DOWN':
                self.sortDirection = 1
            elif sd == 'RANDOM':
                self.sortDirection = random.randint(0, 1)
            else:
                self.sortDirection = 0

            if self.getStrum(sc):
                loo.sort()    # sort for strum only. If no strum it doesn't matter
                if self.sortDirection:
                    loo.reverse()

            # ornametation applies to the top note only. However,
            # we might need to step though the list of notes to find
            # a sounding note (volumes set to 0).
            offset = p.offset
            if self.ornaments['type']:
                while 1:
                    lo = loo.pop()
                    if lo[1]:
                        offset = mma.MMA.ornament.doOrnament(self, [lo],
                                self.getChordInPos(p.offset, ctable).chord.scaleList, p)
                        break
                    elif not loo:
                        break
                    else:
                        continue
            
            # Handle a non-ornamented chord or the remainer of an ornamented one
            self.sendChord(loo, p.duration, offset)

            tb.chord.reset()    # important, other tracks chord object

        # Adjust the voicingMove counter at the end of the bar

        if self.voicing.bcount:
            self.voicing.bcount -= 1



# func.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

Code for Defaults and recursion stack Ignazio Di Napoli.

   Define and execute simple function / subroutine calls.

"""

import copy
from . import gbl
import mma.MMA.file
from mma.MMA.macro import macros
from mma.MMA.common import *
import mma.MMA.debug

# Storage for our functions 
class Funcs:
    def __init__(self, params, body, defaults, fname, lineN):
        self.params = params
        self.body = body
        self.defaults = defaults
        self.lineN = lineN
        self.fname = fname
        
funcList = {}

########################

def defCall(l):
    """ Define a function.  """

    if len(l) < 1:
        error("DefCall: At least one argument (name) is needed.")
        
    lineN = gbl.lineno
        
    # Grab 1st arg. This is the function name
    fname = l[0].strip().upper()
    if '$' in fname:
        error("DefCall: '$' is not permitted in function name.")

    # The params are split into words. Join together just so we can
    # take it apart again. Needed to use ',' delimiters
    # read from the inside out: replace '\,' with chr(255)
    #                           join together with space delims (skip func name)
    #                           split up at ','

    p = ' '.join(l[1:]).replace('\,', chr(255)).split(',')

    # Convert the param names so they have a leading '$'. The '$'
    # is not permitted in a param name.

    params = []
    defaults = []
    for a in p:
        a = a.strip().replace(chr(255), ",")
        if not a:
            continue
        if '=' in a:
            a,d = a.split('=')
            a = a.strip()
            d = d.strip()
        else:
            d = None
        if '$' in a:
            error("DefCall: '$' is not permitted in paramater name.")
        a = '$'+a
        if a in params:
            error("DefCall: '%s' is a duplicate paramater name." % a)

        params.append(a.upper())
        defaults.append(d)
    
    body = []
    while 1:
        ln = gbl.inpath.read()
        if not ln:
            error("DefCall: Reached EOF while looking for EndDefCall.")
        cmd = ln[0].upper()
        if cmd in ("DEFCALLEND", "ENDDEFCALL"):
            if len(ln) > 1:
                error("DefCall: No arguments permitted for DefCallEnd/EndDefCall.")
            else:
                break
        if cmd == "DEFAULT":
            if len(ln) < 3:
                error("Default: Requires two arguments.")
            a = '$'+ln[1].strip().upper()
            d = ' '.join(ln[2:])
            try:
                i = params.index(a)
            except:
                error("DefCall Default: param '%s' does not exist in '%s'." % (a, fname))
            if defaults[i]:
                warning("DefCall Default: param '%s' default value '%s' was set in param list."
                        " Overriding with '%s'." % (a, defaults[i], d))
            defaults[i] = d
            continue
            
        body.append(ln)
    
    funcList[fname] = Funcs(params, body, defaults, gbl.inpath.fname, lineN)
    
    if mma.MMA.debug.debug: 
        t = [ a[1:] for a in params]
        dPrint("DefCall: Created function '%s': %s" % (fname, ', '.join(t)))


def callFunction(l):
    """ Call a function. """

    if len(l) < 1:
        error("DefCall: At least one argument (name) is needed.")

    fname = l[0].strip().upper()
    
    if fname not in funcList:
        error("Call: '%s' has not been defined." % fname)

    # Convert any escaped '\,' into $ff ... see explanation in defCall() 
    p = ' '.join(l[1:]).replace('\,', chr(255)).split(',')

    # Validate calling params
    params = funcList[fname].params

    callParams = []
    usingDefaults = False
    namedParams = {}
    for a in p:
        a = a.strip().replace(chr(255), ',')   # commas restored
        if a:   # new list of params
            if '=' in a: 
                usingDefaults = True
                a,d = a.split('=')
                a = '$'+a.strip().upper()
                d = d.strip()
                if a not in params:
                    error("Call: '%s' is not a param of '%s'." % (a, fname))
                namedParams[a] = d
            elif usingDefaults:
                error("Call: cannot use a non-named param after a named one.")
            else:
                callParams.append(a)
                
    if usingDefaults or len(params) != len(callParams):
        s = len(callParams)
        for i, p in enumerate(params[s:]):
            if p in namedParams:
                callParams.append(namedParams[p])
            else:
                d = funcList[fname].defaults[i+s]
                if d is None:
                    error("Call: '%s' has no default for '%s'." % (fname, p))
                callParams.append(d)

    if len(params) != len(callParams):
           error("Call: Function '%s' needs %s params, '%s' given." % \
                 (fname, len(params), len(callParams)))

    # First, push the current values on the stack.
    # To avoid pushing not existant values, NewSet is called before pushing.
    # Then, set parameters as global values.
    # At the end of the function, stack values are restored.
    # 
    # This way, the parameters become local.  


    existingParams = [name for name in params if name[1:] in macros.vars]
    pushStack = [["StackValue", name] for name in existingParams]
    sets = [["Set", name[1:], value] for name, value in zip(params, callParams)]
    body = funcList[fname].body
    popStack = [["Set", name[1:], "$_StackValue"] for name in existingParams[::-1]]
    unset = [["UnSet", name[1:]] for name in params if name[1:] not in macros.vars]

    fullbody = pushStack + sets + body + popStack + unset
  
    # we use the source line for each, but:
    # - there is no stack for error displaying; error just display current line in input
    # - there is no saving of original file name 
     
    #lineNs = [funcList[fname].lineN] * (len(newSets) + len(pushStack) + len(sets))
    #lineNs += range(funcList[fname].lineN+1, funcList[fname].lineN+1+len(funcList[fname].body))
    #lineNs += [lineNs[-1]+1] * len(popStack)
    ##print zip(lineNs, fullbody)
    #lineNsS = ["{} ({}:{})".format(gbl.lineno, funcList[fname].fname, l) for l in lineNs]
    #gbl.inpath.push(fullbody, lineNsS) 

    # push the converted body lines into the input stream
    gbl.inpath.push(fullbody, [gbl.lineno] * len(fullbody)) 

    #print "=" * 80
    #print fname
    #print zip(params, callParams)
    #print body
    #print "=" * 80

    if mma.MMA.debug.debug:
        dPrint ("Call: function '%s' expanded." % fname)






# chordtable.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>



Table of chords. All are based on a C scale.
Generating chords is easy in MIDI since we just need to
add/subtract constants, based on yet another table.

CAUTION, if you add to this table make sure there are at least
3 notes in each chord! Don't make any chord longer than 8 notes
(The pattern define sets volumes for only 8).

There is a corresponding scale set for each chord. These are
used by bass and scale patterns.

Each chord needs an English doc string. This is extracted by the
-Dn option to print a table of chordnames for inclusion in the
reference manual.

"""

C  = 0
Cs = Db = 1
D       = 2
Ds = Eb = 3
E  = Fb = 4
Es = F  = 5
Fs = Gb = 6
G       = 7
Gs = Ab = 8
A  = Bbb= 9
As = Bb = 10
B  = Cb = 11

chordlist = {
    'M':    ((C,    E,      G ),
             (C, D, E, F, G, A, B),
             "Major triad. This is the default and is used in  "
             "the absence of any other chord type specification."),

    '(b5)':  ((C,    E,      Gb ),
             (C, D, E, F, Gb, A, B),
             "Major triad with flat 5th. MMA notatation requires the () around the name."),

    'add9': ((C,    E,    G,    D+12),
             (C, D, E, F, G, A, D+12),
             "Major chord plus 9th (no 7th.)"),

    'addb9': ((C,    E,    G,    Db+12),
             (C, D, E, F, G, A, Db+12),
             "Major chord plus flat 9th (no 7th.)"),

    'add#9': ((C,    E,    G,    Ds+12),
             (C, D, E, F, G, A, Ds+12),
             "Major chord plus sharp 9th (no 7th.)"),

    'm':    ((C,    Eb,    G ),
             (C, D, Eb, F, G, Ab, Bb),
             "Minor triad."),

    'mb5':    ((C,    Eb,       Gb ),
             (C, D, Eb, F, Gb, Ab, Bb),
             "Minor triad with flat 5th (aka dim)."),

    'm#5':    ((C,    Eb,       Gs ),
             (C, D, Eb, F, Gs, Ab, Bb),
             "Minor triad with augmented 5th."),

    'm6':    ((C,    Eb,       G, A ),
             (C, D, Eb, F, G, A, Bb),
             "Minor 6th (flat 3rd plus a 6th)."),

    'm6(add9)':    ((C, Eb, G, D+12, A+12),
             (C, D, Eb, F, G, A, Bb),
             "Minor 6th with added 9th. This is sometimes notated as a slash chord "
             "in the form ``m6/9''." ),

    'm7':    ((C,    Eb,       G,      Bb ),
             (C, D, Eb, F, G, Ab, Bb),
             "Minor 7th (flat 3rd plus dominant 7th)."),

    'm7#5':    ((C,    E,    Gs,    Bb),
             (C, D, E, F, Gs, A, Bb),
             "Minor 7th with sharp 5th."),

    'mM7':    ((C,    Eb,      G,      B ),
             (C, D, Eb, F, G, Ab, B),
             "Minor Triad plus Major 7th. You will also see this printed "
             "as ``m(maj7)'', ``m+7'', ``min(maj7)'' and ``min$\sharp$7'' "
             "(which \mma\ accepts); as well as the \mma\ \emph{invalid} "
             "forms: ``-($\Delta$7)'', and ``min$\\natural$7''."),

    'm+7b9':  ((C, Eb, Gs, Bb, Db+12),
               (C, Db, Eb, F, Gs, Ab, Bb),
               "Augmented minor 7 plus flat 9th."),

    'm+7#9':  ((C, Eb, Gs, Bb, Ds+12),
               (C, Ds, Eb, F, Gs, Ab, Bb),
               "Augmented minor 7 plus sharp 9th."),
    
    'mM7(add9)': ((C, Eb, G, B, D+12),
                 (C, D, Eb, F, G, Ab, B),
                 "Minor Triad plus Major 7th and 9th."),

    'm7b5': ((C,    Eb,       Gb,       Bb ),
             (C, D, Eb, F, Gb, Ab, Bb),
             "Minor 7th, flat 5 (aka 1/2 diminished). "),

    'm7b9': ((C,     Eb,    G,     Bb, Db+12 ),
             (C, Db, Eb, F, G, Ab, Bb),
             "Minor 7th with added flat 9th."),

    'm7#9': ((C,     Eb,    G,     Bb, Ds+12 ),  # Eb == D# 
             (C, Ds, Eb, F, G, Ab, Bb),
             "Minor 7th with added sharp 9th."),

    'mb9': ((C,    Eb,    G,    Db+12),
             (C, D, Eb, F, G, A, Db+12),
             "Minor chord plus flat 9th (no 7th.)"),

    '7':    ((C,    E,      G,    Bb ),
             (C, D, E, F, G, A, Bb),
             "7th."),

    '7(6)':    ((C,    E,      G, A,   Bb ),
             (C, D, E, F, G, A, Bb),
             "7th with added 6th."),

    '7b5':    ((C,    E,      Gb,    Bb ),
             (C, D, E, F, Gb, A, Bb),
             "7th, flat 5."),

    'dim7': ((C,    Eb,       Gb,       Bbb ),
             (C, D, Eb, F, Gb, Ab, Bbb ),    # missing 8th note
             "Diminished seventh."),
    
    'dim7(addM7)': ((C, Eb, Gb, A, B),
             (C, D, Eb, F, Gb, A, B),
             "Diminished tirad with added Major 7th."),

    'dim(b13)':  ((C, Eb, Gb, Bbb, Ab),
                 (C, D, Eb, F, Gb, Ab, Bbb),
                 "Diminished seventh, added flat 13th."),

    'aug':    ((C,    E,      Gs ),
             (C, D, E, F, Gs, A, B ),
             "Augmented triad."),

    '6':    ((C,    E,      G, A ),
             (C, D, E, F, G, A, B),
             "Major tiad with added 6th."),

    '6(add9)':    ((C,   E, G, D+12, A+12),
             (C, D, E, F, G, A, B),
             "6th with added 9th. This is sometimes notated as a slash chord "
             "in the form ``6/9''. MMA voices the 6th an octave higher."),

    'M7':    ((C,    E,    G,    B),
             (C, D, E, F, G, A, B),
             "Major 7th."),



    'M7#5':    ((C,    E,    Gs,    B),
             (C, D, E, F, Gs, A, B),
             "Major 7th with sharp 5th."),

    'M7b5': ((C,    E,      Gb,    B ),
             (C, D, E, F, Gb, A, B ),
             "Major 7th with a flat 5th."),

    '9':    ((C,    E,    G,    Bb, D+12 ),
             (C, D, E, F, G, A, Bb),
             "7th plus 9th."),

 

    '9b5':    ((C,    E,    Gb,    Bb, D+12 ),
             (C, D, E, F, Gb, A, Bb),
             "7th plus 9th with flat 5th."),

    'm9':    ((C,    Eb,       G,      Bb, D+12 ),
             (C, D, Eb, F, G, Ab, Bb),
             "Minor triad plus 7th and 9th."),

    'm7b5b9': ((C, Eb, Gb, Bb, Db+12),
               (C, Db, Eb, F, Gb, Ab, Bb),
               "Minor 7th with flat 5th and flat 9th."),

    'm9b5': ((C,    Eb,    Gb, Bb, D+12 ),
             (C, D, Eb, F, Gb, Ab, Bb),
             "Minor triad, flat 5, plus 7th and 9th."),

    'm(sus9)':((C,      Eb,    G,     D+12 ),
               (C, D, Eb, F, G, Ab, D+12),
               "Minor triad plus 9th (no 7th)."),

    'M9':    ((C,    E,    G,    B, D+12 ),
             (C, D, E, F, G, A, B),
             "Major 7th plus 9th."),

    'M9#11': ((C, E, G, B, D+12, Fs+12),
              (C, D, E, Fs, G, A, B),
              "Major 9th plus sharp 11th."),

    '7b9':    ((C,     E,    G,    Bb, Db+12 ),
             (C, Db, E, F, G, A, Bb),
             "7th with flat 9th."),

    '7#9':    ((C,     E,    G,    Bb, Ds+12 ),
             (C, Ds, E, F, G, A, Bb),
             "7th with sharp 9th."),

    '7#9b13':    ((C,     E,    G,    Bb, Ds+12, Ab+12 ),
             (C, Ds, E, F, G, Ab, Bb),
             "7th with sharp 9th and flat 13th."),

    '7b5(add13)':    ((C,     E,    Gb,    Bb, A+12 ),
             (C, D, E, F, Gb, A, Bb),
             "7th with flat 5 and 13th."),

    '7(add13)':    ((C,     E,    G,    Bb, A+12 ),
             (C, D, E, F, G, A, Bb),
             "7th with added 13th."),

    '7b5b9':((C,     E,    Gb,    Bb, Db+12 ),
             (C, Db, E, F, Gb, A, Bb),
             "7th with flat 5th and flat 9th."),

    '7b5#9':((C,     E,    Gb,    Bb, Ds+12 ),
             (C, Ds, E, F, Gb, A, Bb),
             "7th with flat 5th and sharp 9th."),

    '7#5#9':((C,     E,    Gs,    Bb, Ds+12 ),
             (C, Ds, E, F, Gs, A, Bb),
             "7th with sharp 5th and sharp 9th."),

 
    'aug7': ((C,    E,    Gs,    Bb ),
             (C, D, E, F, Gs, A, Bb),
             "An augmented chord (raised 5th) with a dominant 7th."),

    'aug7b9':((C,     E,    Gs,    Bb, Db+12 ),
              (C, Db, E, F, Gs, A, Bb),
              "An augmented chord (raised 5th) with a dominant 7th and flat 9th."),

    'aug7#9':((C,     E,    Gs,    Bb, Ds+12 ),
              (C, Ds, E, F, Gs, A, Bb),
              "An augmented chord (raised 5th) with a dominant 7th and sharp 9th."),

    'aug9M7':((C,     E,    Gs,    B, D+12 ),
              (C, D, E, F, Gs, A, B),
              "An augmented chord (raised 5th) with a major 7th and 9th."),

    '+7b9#11': ((C, E, Gs, Bb, Db+12, Fs+12),
                (C, Db, E, Fs, G, A, Bb),
                "Augmented 7th with flat 9th and sharp 11th."),

    'm+7b9#11':  ((C, Eb, Gs, Bb, Db+12, Fs+12),
                (C, Db, Eb, Fs, Gs, A, Bb),
                "Augmented minor 7th with flat 9th and sharp 11th."),

    '11':    ((C,   C,    G,    Bb, D+12, F+12 ),
             (C, D, E, F, G, A, Bb),
             "9th chord plus 11th (3rd not voiced)."),

    '11+':    ((C,   C,    Gs,    Bb, D+12, F+12 ),
             (C, D, E, F, Gs, A, Bb),
             "Augmented 11th (sharp 5)."),
    
    'm11':    ((C,    Eb,    G,     Bb, D+12, F+12 ),
             (C, D, Eb, F, G, Ab, Bb),
             "9th with minor 3rd,  plus 11th."),

    'M11':    ((C,    E,    G,   B, D+12, F+12 ),
             (C, D, E, F, G, Ab, B),
             "Major 9th plus 11th."),

    'm7(add11)':    ((C,    Eb,    G,    Bb, F+12 ),
             (C, D, Eb, F, G, Ab, Bb),
             "Minor 7th  plus 11th."),

    'm9#11':   ((C, Eb, G, Bb, D+12, Fs+12),
                (C, D, Eb, Fs, G, A, Bb),
                "Minor 7th plus 9th and sharp 11th."),

    'm7b9#11':  ((C, Eb, G, Bb, Db+12, Fs+12),
                (C, Db, Eb, Fs, G, A, Bb),
                "Minor 7th plus flat 9th and sharp 11th."),

    'm7(add13)':    ((C,    Eb,    G,    Bb, A+12 ),
             (C, D, Eb, F, G, A, Bb),
             "Minor 7th  plus 13th."),

    '11b9': ((C,     E,    G,    Bb, Db+12, F+12 ),
             (C, Db, E, F, G, A, Bb),
             "7th chord plus flat 9th and 11th."),

    '9#5':    ((C,    E,    Gs,    Bb, D+12 ),
             (C, D, E, F, Gs, A, Bb),
             "7th plus 9th with sharp 5th (same as aug9)."),

    '9b6':   ((C,   E,  Ab, D+12 ),
              (C, D, E, F, Ab, B, D+12),
              "9th with flat 6 (no 5th or 7th)."),
    
    '9#11': ((C,    E,     G,    Bb, D+12, Fs+12 ),
             (C, D, E, Fs, G, A, Bb),
             "7th plus 9th and sharp 11th."),

    '7#9#11':((C,     E,     G,    Bb, Ds+12, Fs+12 ),
              (C, Ds, E, Fs, G, A, Bb),
              "7th plus sharp 9th and sharp 11th."),

    '7b9#11': ((C,     E,     G,    Bb, Db+12, Fs+12 ),
              (C, Db, E, Fs, G, A, Bb),
              "7th plus flat 9th and sharp 11th."),

    '7#11':((C,    E,     G,    Bb,  Fs+12 ),
             (C, D, E, Fs, G, A, Bb),
             "7th plus sharp 11th (9th omitted)."),

    'M7#11':((C,    E,     G,    B,  Fs+12 ),
             (C, D, E, Fs, G, A, B),
             "Major 7th plus sharp 11th (9th omitted)."),

    'm11b5': ((C, Eb, Gb, Bb, D+12, F+12),
              (C, D, Eb, F, Gb, A, Bb),
              "Minor 7th with flat 5th plus 11th."),

    # Sus chords. Not sure what to do with the associated scales. For
    # now just duplicating the 2nd or 3rd in the scale seems to make sense.

    'sus4': ((C,    F,    G ),
             (C, D, F, F, G, A, B),
             "Suspended 4th, major triad with the 3rd raised half tone."),

    'msus4':   ((C,    Eb,  F,  G ),
             (C, D, Eb, F, G, Ab, Bb),
             "Minor suspended 4th, minor triad plus 4th."),

    'm7sus4':  ((C,    Eb,  F,  G, Bb ),
             (C, D, Eb, F, G, Ab, Bb),
             "Minor suspended 4th, minor triad plus 4th and dominant 7th."),

    'sus(addb9)': ((C,    F,    G, Db+12 ),
             (C, D, F, F, G, A, B),
             "Suspended 4th, major triad with the 3rd raised half tone plus flat 9th."),

    'sus(add9)': ((C,    F,    G,  D+12 ),
             (C, D, F, F, G, A, B),
             "Suspended 4th, major triad with the 3rd raised half tone plus 9th."),

    'sus(add#9)': ((C,    F,    G, Ds+12 ),
             (C, D, F, F, G, A, B),
             "Suspended 4th, major triad with the 3rd raised half tone plus sharp 9th."),

    '7sus': ((C,    F,    G,    Bb ),
             (C, D, F, F, G, A, Bb),
             "7th with suspended 4th, dominant 7th with 3rd "
             "raised half tone."),

    '7susb9': ((C, F, G, Bb, Db+12),
               (C, Db, F, F, G, A, Bb),
               "7th with suspended 4th and flat 9th."),

    'sus2': ((C,    D,    G ),
             (C, D, D, F, G, A, B),
             "Suspended 2nd, major triad with the major 2nd above the "
             "root substituted for 3rd."),

    '7sus2':((C,    D,    G,    Bb ),
             (C, D, D, F, G, A, Bb),
             "A sus2 with dominant 7th added."),

    'sus9': ((C,    F,    G,    Bb, D+12),
             (C, D, F, F, G, A, Bb),
             "7sus plus 9th."),

    '13sus': ((C, F, G, Bb, D+12, A+12),
               (C, D, F, F, G, A, Bb),
               "7sus, plus 9th and 13th"),

    '13susb9': ((C, F, G, Bb, Db+12, A+12),
               (C, Db, F, F, G, A, Bb),
               "7sus, plus flat 9th and 13th"),

    # these chords should probably NOT have the 5th included,
    # but since a number of voicings depend on the 5th being
    # the third note of the chord, they're here.

    '13':    ((C,    E,    G,    Bb, A+12),
             (C, D, E, F, G, A, Bb),
             "7th (including 5th) plus 13th (the 9th and 11th are not voiced)."),

    '7b13':  ((C,    E,    G,    Bb, Ab+12),
             (C, D, E, F, G, A, Bb),
             "7th (including 5th) plus flat 13th (the 9th and 11th are not voiced)."),

    '13b5':  ((C,    E,    Gb,    Bb, Ab+12),
             (C, D, E, F, Gb, A, Bb),
             "7th with flat 5th,  plus 13th (the 9th and 11th are not voiced)."),

    '13#9':    ((C,    E,    G,    Bb, Ds+12,  A+12),
             (C, Ds, E, F, G, A, Bb),
             "7th (including 5th) plus 13th and sharp 9th (11th not voiced)."),
 
    '13b9':  ((C,    E,    G,    Bb, Db+12,  A+12),
             (C, Db, E, F, G, A, Bb),
             "7th (including 5th) plus 13th and flat 9th (11th not voiced)."),
    
    'M#11':   ((C,    E,    G,    B, Fs+12),
             (C, D, E, Fs, G, A, B),
             "Major triad plus sharp 11th."),
    
    'M13':   ((C,    E,    G,    B, A+12),
             (C, D, E, F, G, A, B),
             "Major 7th (including 5th) plus 13th (9th and  11th not voiced)."),

    'm13':   ((C, Eb, G, Bb, A+12),
              (C, D, Eb, F, G, A, Bb),
              "Minor 7th (including 5th) plus 13th (9th and 11th not voiced)."),

    '13#11': ((C,    E,    G,    Bb, Fs+12, A+12),
             (C, D, E, Fs, G, A, Bb),
             "7th plus sharp 11th and 13th (9th not voiced)."),

    'M13#11': ((C,    E,    G,    B, Fs+12, A+12),
             (C, D, E, Fs, G, A, B),
             "Major 7th plus sharp 11th and 13th (9th not voiced)."),

    # Because some patterns assume that the 3rd note in a chord is a 5th,
    # or a varient, we duplicate the root into the position of the 3rd ... and
    # to make the sound even we duplicate the 5th into the 4th position as well.

    '5':    ((C, C,    G, G ),
             (C, D, E, F, G, A, B),
             "Altered Fifth or Power Chord; root and 5th only."),

    'omit3add9': ((C, C, G, D+12),
               (C, D, E, F, G, A, Bb),
               "Triad: root, 5th and 9th."),

    '7omit3': ((C, C, G, Bb),
                (C, D, E, F, G, A, Bb),
                "7th with unvoiced 3rd."),

    'm7omit5': ((C, Eb, Bb),
                 (C, D, Eb, F, G, A, Bb),
                 "Minor 7th with unvoiced 5th."),
}


# Extend our table with common synomyns. These are real copies,
#   not pointers. This is done so that a user redefine only affects
#   the original. There is no way to tell if a chordlist[] entry is
#   from above or if it's an added alias or defChord.

aliases = (
    ('11#5',     '11+',      ''),
    ('aug9',     '9#5',      ''),
    ('9+',       '9#5',      ''),
    ('+9',       '9#5',      ''),
    ('+9M7',     'aug9M7',   ''),
    ('+M7',      'M7#5',     ''),
    ('m(add9)',  'm(sus9)',  ''),
    ('(add9)',   'add9',     ''),
    ('(addb9)',  'addb9',    ''),
    ('(add#9)',  'add#9',    ''),
    ('69',       '6(add9)',  ''),
    ('m69',      'm6(add9)', ''),
    ('m(b5)',    'mb5',      ''),
    ('m7(b9)',   'm7b9',     ''),
    ('m7(#9)',   'm7#9',     ''),
    ('9+5',      '9#5',      ''),
    ('m+5',      'm#5',      ''),
    ('m+',       'm#5',      ''),
    ('M6',       '6',        ''),
    ('m7-5',     'm7b5',     ''),
    ('m7(omit5)','m7omit5',  ''),
    ('+',        'aug',      ''),
    ('+7',       'aug7',     ''),
    ('+7#9',     'aug7#9',   ''),
    ('+7b9',     'aug7b9',   ''),
    ('7(omit3)', '7omit3',   ''),
    ('(#5)',     'aug',      ''),
    ('7#5b9',    'aug7b9',   ''),
    ('7-9',      '7b9',      ''),
    ('7+9',      '7#9',      ''),
    ('maj7',     'M7',       ''),
    ('M7-5',     'M7b5',     ''),
    ('M7+5',     'M7#5',     ''),
    ('M7(add13)','13b9',     ''),
    ('7alt',     '7b5b9',    'Uses a 7th flat 5, flat 9. Probably not correct, but works (mostly).'),
    ('7sus4',    '7sus',     ''),
    ('13sus4',   '13sus',    ''),
    ('7+',       'aug7',     ''),
    ('7#5',      'aug7',     ''),
    ('7+5',      'aug7',     ''),
    ('7-5',      '7b5',      ''),
    ('sus',      'sus4',     ''),
    ('maj9',     'M9',       ''),
    ('maj13',    'M13',      ''),
    ('m(maj7)',  'mM7',      ''),
    ('m+7',      'mM7',      ''),
    ('min(maj7)','mM7',      ''),
    ('min#7',    'mM7',      ''),
    ('m#7',      'mM7',      ''),
    ('msus',     'msus4',    ''),
    ('m7sus',    'm7sus4',   ''),
    ('dim',      'dim7',     'A dim7, not a triad!'),
    (chr(176),   'dim7',     'A dim7 using a degree symbol'),
    (chr(176)+'3', 'mb5',   'A dim3 (triad) using a degree symbol'),
    (chr(176)+'(addM7)', 'dim7(addM7)', 'dim7(addM7) using degree symbol'),
    (chr(248),   'm7b5',     'Half-diminished using slashed degree symbol'),
    ('9sus',     'sus9',     ''),
    ('9-5',      '9b5',      ''),
    ('dim3',     'mb5',      'Diminished triad (non-standard notation).'),
    ('omit3(add9)','omit3add9', ''),
    ('9sus4',    'sus9',     ''),
    ('7b9sus',   '7susb9',   '')
    )

for a, b, d in aliases:
    n = chordlist[b][0]
    s = chordlist[b][1]
    if not d:
        d = chordlist[b][2]

    chordlist[a] = (n, s, d)

### little snippet to print chord info for my accordion chord program.
##print "chords = {"
##for a in chordlist:
##    print "   '%s': ( %s,  \"%s\")," % (a, chordlist[a][0], chordlist[a][2])
##print "}"



# midinote.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

This module does all the midinote stuff.

"""


import mma.MMA.notelen
import mma.MMA.midiC
from   mma.MMA.midiM import packBytes
import mma.MMA.debug
from . import gbl
from   mma.MMA.common import *
from   mma.MMA.keysig import keySig


def parse(name, ln):
    """ Called from parser for a <Track MidiNote> This figures the right routine."""

    if not len(ln):
        error("MidiNote: Needs arguments")

    trk = gbl.tnames[name]

    # parse out the cmd=value options pairs

    ln, opts = opt2pair(ln)

    for o, v in opts:
        o = o.upper()
        v = v.upper()
        if o == 'TRANSPOSE':
            if v in ('0', 'OFF'):
                trk.transpose = 0
            elif v in ('1', 'ON'):
                trk.transpose = gbl.transpose
            else:
                error("MIDINote: TRANSPOSE expecting ON or OFF, not %s." % v)

        elif o == 'OFFSETS':
            if v == 'BEATS':
                trk.useticks = 0
            elif v == 'TICKS':
                trk.useticks = 1
            else:
                error("MIDINote: OFFSETS expecting BEATS or TICKS, not %s." % v)

        elif o == 'DURATION':
            if v == 'NOTES':
                trk.tickdur = 0
            elif v == 'TICKS':
                trk.tickdur = 1
            else:
                error("MIDINote: DURATION expecting NOTES or TICKS, not %s." % v)

        elif o == 'ARTICULATE':
            if v in ('1,' 'ON'):
                trk.articulate = 1
            elif v in ('0', 'OFF'):
                trk.articulate = 0
            else:
                error("MIDINote: ARTICULATE expecting ON or OFF, not %s." % v)

        elif o == 'OCTAVE':
            trk.oadjust = stoi(v)
            if trk.oadjust < -4 or trk.oadjust > 4:
                error("MIDINote: Octave adjustment must be -4..4, not '%s'." % trk.oadjust)

        elif o == 'VOLUME':
            trk.vadjust = stof(v) / 100
            if trk.vadjust <= 0:
                error("MIDINote: Volume %% adjustment must be > 0, not '%s'." % trk.vadjust)

        elif o == 'ADJUST':
            trk.tadjust = stoi(v)

        else:
            error("MIDINote: unknown option pair %s=%s." % (o, v))
    
    # end of option pairs.

    if ln:  # process rest of the stuff ... real midi events.

        trk.setForceOut()

        a = ln[0].upper()

        if a[0] in (".0123456789"):
            insertNote(trk, ln)

        elif a == "NOTE":
            insertNote(trk, ln[1:])

        elif a == "PB":
            insertPB(trk, ln[1:])

        elif a == "PBR":
            insertPBrange(trk, ln[1:])

        elif a == "CTRL":
            insertControl(trk, ln[1:])

        elif a == "CHAT":
            insertChTouch(trk, ln[1:])

        elif a == "CHATR":
            insertChTouchRange(trk, ln[1:])

        else:
            error("MidiNote: Unknown command '%s'." % a)

    if mma.MMA.debug.debug:
        if opts:
            dPrint("MIDINOTE: %s" % mopts(trk))


def mopts(trk):
    """ Return options string to macro and setoption-debug. """

    if trk.useticks:
        a1 = 'Ticks'
    else:
        a1 = 'Beats'

    if trk.tickdur:
        a2 = 'Ticks';
    else:
        a2 = 'Notes'

    if trk.transpose:
        a3 = 'On'
    else:
        a3 = 'Off'

    if trk.articulate:
        a4 = 'On'
    else:
        a4 = 'Off'

    return "Offsets=%s Duration=%s Transpose=%s Articulate=%s Adjust=%s Volume=%s Octave=%s" \
        % (a1, a2, a3, a4, trk.tadjust, trk.vadjust * 100, trk.oadjust)


def getoffset(trk, v):
    """ Convert a string (value) to an offset. Convert to beats if nesc."""

    offset = stof(v)
    if trk.useticks:
        if offset != int(offset):
            error("MidiNote: Offset set to ticks, float '%s' given. Integer must be used." %
                  offset)
        offset += trk.tadjust

    else:
        if offset < 1:
            warning("MidiNote: Offset %s generates notes before start of current bar." % offset)

        if offset >= gbl.QperBar + 1:
            warning("MidiNote: Offset %s is past end of current bar." % offset)

        offset = (offset-1) * gbl.BperQ

    return int(offset)  # conversion to int is nescessary


def note2val(trk, acctable, orig):
    """ Convert a note name to a value. In this case the OCTAVE setting
        is used!
    """

    t = list(orig)
    n = t.pop(0)
    try:
        val = {'c': 0, 'd': 2, 'e': 4, 'f': 5, 'g': 7, 'a': 9, 'b': 11}[n]
    except:
        error("MidiNote: Expecting valid note name, not '%s'." % orig)

    val += trk.octave[gbl.seqCount]  # add in current octave

    # Modify the note with either the keysignature or given accidental.

    if t:   # override modifier if #,& or n
        if t[0] == '#':
            acctable[n] = 1
            t.pop(0)
        elif t[0] == '&':
            acctable[n] = -1
            t.pop(0)
        elif t[0] == 'n':
            acctable[n] = 0
            t.pop(0)

    val += acctable[n]

    # adjust for octave - or +

    while t and (t[0] == '-' or t[0] == '+'):
        if t[0] == '+':
            val += 12
        else:
            val -= 12
        t.pop(0)

    if t:  # anything left? Error.
        error("MidiNote: Unknown note specifier '%s' in '%s'." % (''.join(t), orig))

    return val


def insertNote(trk, ln):
    """ Insert specified (raw) MIDI notes into track. """

    if len(ln) != 4:
        error("Use: %s MidiNote: <offset> <note> <velocity> <duration>" % trk.name)

    acctable = keySig.accList   # keysig modifier, use for chord

    offset = getoffset(trk, ln[0])

    # Set a flag if this is a drum track.

    if trk.vtype == 'DRUM':
        isdrum = 1
    elif trk.vtype in ('MELODY', 'SOLO') and trk.drumType:
        isdrum = 1
    else:
        isdrum = 0

    notes = []
    for n in ln[1].split(','):
        if n[0] in '0123456789':
            n = stoi(n)
        else:
            if isdrum:
                if n == '*':
                    if trk.vtype in ('MELODY', 'SOLO'):
                        n = trk.drumTone
                    else:
                        n = trk.toneList[gbl.seqCount]
                else:
                    n = mma.MMA.midiC.drumToValue(n)
                    if n < 0:
                        error("MidiNote: unknown drum tone '%s' in %s." % (n, trk.name))
            else:
                n = note2val(trk, acctable, n)

        if n < 0 or n > 127:
            error("MidiNote: Notes must be in the range 0...127, not %s" % n)

        if trk.transpose and not isdrum:
            n += gbl.transpose
            while n < 0:
                n += 12
            while n > 127:
                n -= 12

        if trk.oadjust and not isdrum:
            n += (trk.oadjust * 12)
            while n < 0:
                n += 12
            while n > 127:
                n -= 12

        notes.append(n)

    velocity = stoi(ln[2])

    if velocity < 1 or velocity > 127:
        error("MidiNote: Note velocity must be in the range 1...127, not %s" % velocity)

    velocity *= trk.vadjust

    if velocity < 1:
        velocity = 1
    elif velocity > 127:
        velocity = 127

    velocity = int(velocity)   # trk.adjust can be a float

    if trk.tickdur:
        duration = stoi(ln[3])
    else:
        duration = mma.MMA.notelen.getNoteLen(ln[3])
        if trk.articulate:
            duration = (duration * trk.artic[gbl.seqCount]) // 100
            if duration < 1:
                duration = 1

    channel = trk.channel
    track = gbl.mtrks[channel]

    for n in notes:
        onEvent = packBytes((0x90 | channel-1, n, velocity))
        offEvent = packBytes(onEvent[:-1], 0)

        track.addToTrack(gbl.tickOffset + offset, onEvent)
        track.addToTrack(gbl.tickOffset + offset + duration, offEvent)

    if mma.MMA.debug.debug:
        dPrint("MidiNote Note %s: inserted note %s at offset %s." % (trk.name, notes, offset))


def insertPB(trk, ln):
    """ Insert a pitch controller event. """

    if len(ln) != 2:
        error("MidiNote: PB expecting 2 arguments.")

    offset = getoffset(trk, ln[0])
    v = stoi(ln[1])

    if v < -8192 or v > 8191:
        error("MidiNote: PB value must be -8192..+8191, not '%s'." % v)

    v += 8192  # convert to 0..16383, max 14 bit value

    channel = trk.channel
    track = gbl.mtrks[channel]
    track.addToTrack(gbl.tickOffset + offset, packBytes((0xe0 | channel-1, v % 128, v // 128)))

    if mma.MMA.debug.debug:
        dPrint("MidiNote PB %s: inserted bend %s at offset %s." % (trk.name, v-8192, offset))


def insertPBrange(trk, ln):
    """ Insert a range of PB events. """

    if len(ln) != 3:
        error("MidiNote: PBR expecting 3 arguments <count> <start,end> <v1,v2>.")

    count = stoi(ln[0])
    try:
        s1, s2 = ln[1].split(',')
    except:
        error("MidiNote PBR: event range must be 'v1,v2', not '%s'." % ln[1])
    s1 = getoffset(trk, s1)
    s2 = getoffset(trk, s2)
    tinc = (s2-s1) / float(count)

    try:
        v1, v2 = ln[2].split(',')
    except:
        error("MidiNote PBR: pitch blend range must be 'v1,v2', not '%s'." % ln[2])
    v1 = stoi(v1)
    v2 = stoi(v2)

    if v1 < -8192 or v1 > 8191 or v2 < -8192 or v2 > 8191:
        error("MidiNote: PBR values must be -8192..+8191, not '%s'." % ln[2])

    v1 += 8192  # convert to 0..16383, max 14 bit value
    v2 += 8192
    vinc = (v2-v1) / float(count)

    channel = trk.channel
    track = gbl.mtrks[channel]

    ev = packBytes(0xe0 | channel-1)
    offset = s1
    bend = v1
    for i in range(count+1):
        v = int(bend)
        track.addToTrack(gbl.tickOffset + int(offset), packBytes(ev, v % 128, v // 128))
        offset += tinc
        bend += vinc

    if mma.MMA.debug.debug:
        dPrint("MidiNote PBR %s: inserted bends %s to %s at offsets %s to %s." % 
            (trk.name, v1-8192, v2-8192, s1, s2))


def insertControl(trk, ln):
    """ Insert a controller event. """

    if len(ln) != 3:
        error("MidiNote: Controller expecting 3 arguments.")

    offset = getoffset(trk, ln[0])

    v = mma.MMA.midiC.ctrlToValue(ln[1])
    if v < 0:
        v = stoi(ln[1])
        if v < 0 or v > 0x7f:
            error("MidiNote: Controller values must be 0x00 to 0x7f, not '%s'." % ln[1])

    d = stoi(ln[2])
    if d < 0 or d > 0x7f:
        error("MidiNote: Control Datum value must be 0x00 to 0x7f, not '%s'." % ln[2])

    channel = trk.channel
    track = gbl.mtrks[channel]

    # bypass the addctl() defined in midi.py just to keep all the calls in this
    # module similar. We should have add**() command in midi.py for the above stuff
    # and redo this.???

    track.addToTrack(gbl.tickOffset + offset, packBytes((0xb0 | channel-1, v, d)) )

    if mma.MMA.debug.debug:
        dPrint("MidiNote Ctrl %s: inserted Controller %s value %s at offset %s." % 
            (trk.name, v, d, offset))


def insertChTouch(trk, ln):
    """ Insert a channel aftertouch) event. """

    if len(ln) != 2:
        error("MidiNote: ChAT expecting 2 arguments.")

    offset = getoffset(trk, ln[0])

    v = stoi(ln[1])

    if v < 0 or v > 127:
        error("MidiNote: ChAT value must be 0 .. 127, not '%s'." % v)

    channel = trk.channel
    track = gbl.mtrks[channel]

    track.addToTrack(gbl.tickOffset + offset, packBytes((0xd0 | channel-1, v)))
 
    if mma.MMA.debug.debug:
        dPrint("MidiNote ChAT %s: inserted channel aftertouch %s at offset %s." % 
            (trk.name, v, offset))


def insertChTouchRange(trk, ln):
    """ Insert a range of channel aftertouch events. """

    if len(ln) != 3:
        error("MidiNote: ChATR expecting 3 arguments <count> <start,end> <v1,v2>.")

    count = stoi(ln[0])
    try:
        s1, s2 = ln[1].split(',')
    except:
        error("MidiNote ChATR: event range must be 'v1,v2', not '%s'." % ln[1])
    s1 = getoffset(trk, s1)
    s2 = getoffset(trk, s2)
    tinc = (s2-s1) / float(count)

    try:
        v1, v2 = ln[2].split(',')
    except:
        error("MidiNote ChATR: range must be 'v1,v2', not '%s'." % ln[2])
    v1 = stoi(v1)
    v2 = stoi(v2)

    if v1 < 0 or v1 > 127 or v2 < 0 or v2 > 127:
        error("MidiNote: ChATR values must be 0.. 127, not '%s'." % ln[2])

    vinc = (v2-v1) / float(count)

    channel = trk.channel
    track = gbl.mtrks[channel]

    ev = packBytes(0xd0 | channel-1)
    offset = s1
    bend = v1
    for i in range(count+1):
        v = int(bend)
        track.addToTrack(gbl.tickOffset + int(offset), packBytes(ev, v))
        offset += tinc
        bend += vinc

    if mma.MMA.debug.debug:
        dPrint("MidiNote ChATR %s: inserted events %s to %s at offsets %s to %s." %
            (trk.name, v1, v2, s1, s2))



# patDrum.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import mma.MMA.notelen
import mma.MMA.translate

from . import gbl
from mma.MMA.common import *
from mma.MMA.pat import PC, Pgroup


class Drum(PC):
    """ Pattern class for a drum track. """

    vtype = 'DRUM'

    def __init__(self, ln):
        """ init for drum track. """

        self.toneList = [38]

        PC.__init__(self, ln)   # This order is important!

        self.setChannel('10')
        if not gbl.mtrks[self.channel].trackname:
            gbl.mtrks[self.channel].addTrkName(0, 'Drum')

    def saveGroove(self, gname):
        """ Save special/local variables for groove. """

        PC.saveGroove(self, gname)  # do this 1st. Creates storage.
        self.grooves[gname]['TONES'] = self.toneList[:]

    def restoreGroove(self, gname):
        """ Restore special/local/variables for groove. """

        self.toneList = self.grooves[gname]['TONES']
        PC.restoreGroove(self, gname)

    def setSeqSize(self):
        """ Expand existing pattern list. """

        self.toneList = seqBump(self.toneList)
        PC.setSeqSize(self)

    def clearSequence(self):
        """ Set some initial values. Called from init and clear seq. """

        PC.clearSequence(self)
        self.toneList = seqBump([38])

    def setTone(self, ln):
        """ Set a tone list. Only valid for DRUMs.
        ln[] is not nesc. the right length.
        """

        ln = lnExpand(ln, '%s Tone' % self.name)
        tmp = []

        for n in ln:
            tmp.append(mma.MMA.translate.dtable.get(n))

        self.toneList = seqBump(tmp)

    def restart(self):
        self.ssvoice = -1

    def getPgroup(self, ev):
        """ Get group for a drum pattern.

            Fields - start, length, volume
        """

        if len(ev) != 3:
            error("There must be at exactly 3 items in each "
                  "group of a drum define, not <%s>" % ' '.join(ev))

        a = Pgroup()

        a.offset = self.setBarOffset(ev[0])
        a.duration = mma.MMA.notelen.getNoteLen(ev[1])
        a.vol = stoi(ev[2], "Type error in Drum volume")

        return a

    def trackBar(self, pattern, ctable):
        """ Do a drum bar.

        Called from self.bar()

        """

        
        sc = self.seq

        for p in pattern:
            tb = self.getChordInPos(p.offset, ctable)
            if tb.drumZ:
                continue

            self.sendNote(
                p.offset,
                self.getDur(p.duration),
                self.toneList[sc],
                self.adjustVolume(p.vol, p.offset))



# patch.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

This module contains the various patch manger routines.

"""

from mma.MMA.common import *
from mma.MMA.miditables import *
import mma.MMA.file
import mma.MMA.midiC


def patch(ln):
    """ Main routine to manage midi patch names. """

    for i, a in enumerate(ln):

        if a.count('=') == 1:
            a, b = a.split('=')
        else:
            b = ''
        a = a.upper()

        if a == "LIST":
            b = b.upper()
            if b == "ALL":
                plistall()
            elif b == "EXT":
                plistext()
            elif b == "GM":
                plistgm()
            else:
                error("Expecting All, EXT or GM argument for List")

        elif a == "RENAME":
            prename(ln[i + 1:])
            break

        elif a == 'SET':
            patchset(ln[i + 1:])
            break

        else:
            error("Unknown option for Patch: %s" % a)

# Set a patch value=name


def patchset(ln):
    if not ln:
        error("Patch Set expecting list of value pairs.")

    notopt, ln = opt2pair(ln, toupper=True)
    if notopt:
        error("PATCH SET: All options must be VALUE=PAIR items.")

    for v, n in ln:
        v = v.split('.')
        if len(v) > 3 or len(v) < 1:
            error("Patch Set: Expecting a voice value Prog.MSB.LSB.")

        # Parse out value. Can be 'nn', 'nn.nn' or 'nn.nn.nn'
        # the parts are for MSB(ctrl32), LSB(ctrl0), and Patch

        voc = 0
        if len(v) > 2:    # ctrl32
            i = stoi(v[2], "Patch Set LSB expecting integer.")
            if i < 0 or i > 127:
                error("LSB must be 0..127, not '%s'." % i)
            voc = i << 16

        if len(v) > 1:    # ctrl0
            i = stoi(v[1], "Patch Set MSB expecting integer.")
            if i < 0 or i > 127:
                error("MSB must be 0..127, not '%s'." % i)
            voc += i << 8

        i = stoi(v[0], "Patch Set Voice expecting integer.")
        if i < 0 or i > 127:
            error("Program must be 0..127, not '%s'." % i)
        voc += i

        # Handle the name.

        if voc in voiceNames:
            warning("Patch Set duplicating voice name %s=%s with %s=%s" %
                    (voiceNames[voc], voc, n, mma.MMA.midiC.extVocStr(voc)))

        if n in voiceInx:
            warning("Patch Set duplicating voice value %s=%s with %s=%s" %
                    (mma.MMA.midiC.extVocStr(voiceInx[n]), n, mma.MMA.midiC.extVocStr(voc), n))

        voiceNames[voc] = n
        voiceInx[n] = voc

# Rename


def prename(ln):
    if not ln:
        error("Patch Rename expecting list of value pairs.")

    notopt, ln = opt2pair(ln, 1)

    if notopt:
        error("PATCH RENAME: expecting OLDNAME=NEWNAME pairs.")

    for a, b in ln:
        if not a in voiceInx:
            error("PATCH RENAME: oldname '%s' doen't exist, can't be renamed." % a)

        if b in voiceInx:
            error("PATCH RENAME: newname '%s' already exists." % b)

        v = voiceInx[a]
        voiceNames[v] = b
        del voiceInx[a]
        voiceInx[b] = v

# list funcs


def plistgm():
    """ List GM voices. """

    for v in sorted(voiceNames.keys()):
        if v <= 127:
            print("%s=%s" % (mma.MMA.midiC.extVocStr(v), voiceNames[v]))


def plistext():
    """ List extended voices. """

    for v in sorted(voiceNames.keys()):
        if v > 127:
            print("%s=%s" % (mma.MMA.midiC.extVocStr(v), voiceNames[v]))


def plistall():
    """ List all voices. """
    plistgm()
    plistext()



# swing.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>


"""

from . import gbl
from mma.MMA.common import *
import mma.MMA.debug
from mma.MMA.notelen import noteLenTable

mode = 0      # defaults to 0, set to 1 for swing mode
skew = None   # this is just for $_SwingMode macro
accent1 = 1      # velocity % adjustments for 1st/2nd notes of swing pattern
accent2 = 1
delay1 = 0      # value, in ticks, for additional delay
delay2 = 0
noteValue = 8    # this can be 8 or 16 for 8th swing or 16th swing

# These 2 funcs are called by the groove save/restore funcs. They
# are used just to make the groove code a bit more readable.


def gsettings():
    return (mode, skew, accent1, accent2, delay1, delay2, noteLenTable['81'], noteLenTable['82'], noteValue)


def grestore(s):
    global mode, skew, accent1, accent2, delay1, delay2, noteValue

    mode, skew, accent1, accent2, delay1, delay2, \
        noteLenTable['81'], noteLenTable['82'], noteValue = s


def swingMode(ln):
    """ Enable/Disable Swing timing mode. """

    global skew, mode, accent1, accent2, delay1, delay2, noteValue

    emsg = "Use: SwingMode [ ON | 1, OFF | 0, SKEW=nn | Accent=nn,nn | Delay=nn,nn | Notes=8|16 ]."

    if not ln:
        error(emsg)

    ln, opts = opt2pair(ln, toupper=True)

    for v in ln:
        if v in ("ON", "1"):
            mode = 1
            continue

        elif v in ("OFF", "0"):
            mode = 0
            continue

        else:
            error(emsg)

    for v, o in opts:
        if v == 'SKEW':
            skew = o
            a = int(stoi(o) * gbl.BperQ / 100)
            noteLenTable['81'] = a
            noteLenTable['82'] = gbl.BperQ - a

        elif v == 'ACCENT':
            if o.count(',') != 1:
                error("Swingmode: ACCENT expecting comma separated values, not '%s'." % o)
            a1, a2 = o.split(',')
            a1 = stoi(a1)
            a2 = stoi(a2)
            if a1 < 1 or a1 > 200 or a2 < 1 or a2 > 200:
                error("Swingmode: Both ACCENT values must be 1..200, not %s,%s."
                      % (a1, a2))

            accent1 = a1 / 100.
            accent2 = a2 / 100.

        elif v == 'DELAY':
            if o.count(',') != 1:
                error("Swingmode: DELAY expecting comma separated values, not '%s'." % o)
            a1, a2 = o.split(',')
            a1 = stoi(a1)
            a2 = stoi(a2)
            if a1 < -20 or a1 > 20 or a2 < -20 or a2 > 20:
                error("Swingmode: Both DELAY values must be -20..20, not %s,%s."
                      % (a1, a2))

            delay1 = a1
            delay2 = a2

        elif v == 'NOTES':
            o = stoi(o)

            if o not in (8, 16):
                error("Swingmode: NOTES expecting 8 or 16, not '%s'." % o)

            noteValue = o

        else:
            error(emsg)

    if mma.MMA.debug.debug:
        dPrint("SwingMode: Status=%s; Accent=%s,%s; Delay=%s,%s; Skew Note lengths: " 
            "%s and %s ticks. Notes=%s" % 
            (mode, int(accent1 * 100), int(accent2 * 100), delay1, delay2,
             noteLenTable['81'], noteLenTable['82'], noteValue))


def settings():
    """ Return string of current settings. For macros. """

    if mode:
        a = "On"
    else:
        a = "Off"

    return "%s Accent=%s,%s Delay=%s,%s Skew=%s Notes=%s" % \
        (a, int(accent1 * 100), int(accent2 * 100), delay1, delay2, skew, noteValue)


def getBeats():
    """ Calc on and off beats for swing. This will work if "notevalue" is
        set to 8 or 16.
    """

    len8 = noteLenTable['8']
    len81 = noteLenTable['81']
    len82 = noteLenTable['82']

    rng = int(gbl.QperBar)
    cnt = gbl.BperQ

    if noteValue == 16:
        len8 /= 2
        len81 /= 2
        len82 /= 2
        rng *= 2
        cnt /= 2

    onBeats = [x * cnt for x in range(rng)]
    offBeats = [(x * cnt + len8) for x in range(rng)]

    return (len8, len81, len82, onBeats, offBeats)


def pattern(plist, vtype):
    """ Do swing adjustments for pattern defs. """

    len8, len81, len82, onBeats, offBeats = getBeats()

    for p in plist:
        if p.duration == len8 or (vtype == "DRUM" and p.duration == 1):
            if p.offset in onBeats:
                if p.duration == len8:
                    p.duration = len81
                    adj = accent1
                    if isinstance(p.vol, list):
                        p.vol = [x * adj for x in p.vol]
                    else:
                        p.vol *= adj
                    p.offset += delay1

            elif p.offset in offBeats:
                if p.duration == len8:
                    p.duration = len82
                    adj = accent2
                    if isinstance(p.vol, list):
                        p.vol = [x * adj for x in p.vol]
                    else:
                        p.vol *= adj
                i = offBeats.index(p.offset)
                p.offset = onBeats[i] + len81 + delay2

    return plist


def swingSolo(notes):
    """ Adjust an entire bar of solo note chords for swingmode.

        Check each chord in the array of chords for a bar for
        successive 8ths on & off the beat. If found, the first is
        converted to 'long' 8th, the 2nd to a 'short'
        and the offset for the 2nd is adjusted to comp. for the 'long'.

        If there is a spurious offset between an on/off beat that pair
        will NOT be adjusted. Not sure if that is right or not?

        Only called from getLine(), this is a separate function for sanity.
    """

    len8, len81, len82, onBeats, offBeats = getBeats()

    nl = sorted(notes)   # list of offsets

    for i in range(len(nl) - 1):

        # Check for successive note event offsets on 8th note positions

        if nl[i] in onBeats and nl[i + 1] == nl[i] + len8:
            beat0 = nl[i]
            beat1 = nl[i + 1]

            # check that all notes are 8ths by comparing a set of all
            # the durations in both offsets with set([len8])

            if set([nev.duration for nev in notes[beat0] + notes[beat1]]) == set([len8]):

                # lengthen notes on-the-beat

                for nev in notes[beat0]:
                    nev.duration = len81
                    nev.velocity *= accent1
                    nev.defvelocity *= accent1

                # if we have a delay for the first note we push
                # the whole array by the correct value.

                if delay1:
                    notes[beat0 + delay1] = notes[beat0]
                    del notes[beat0]

                # shorten notes off-the-beat

                for nev in notes[beat1]:
                    nev.duration = len82
                    nev.velocity *= accent2
                    nev.defvelocity *= accent2

                # move notes off-the-beat to the proper offset.
                src = beat1
                dest = beat0 + len81 + delay2
                if src != dest:
                    notes[dest] = notes[src]
                    del notes[src]

    return notes


# This forces our skew value default and in the process
# it sets the durations for the 81/82 notes

swingMode(['Skew=66'])





"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

from mma.MMA.common import *
import mma.MMA.debug
import re

""" Volumes are specified in musical terms, but converted to
    midi velocities. This table has a list of percentage changes
    to apply to the current volume. Used in both track and global
    situations. Note that the volume for 'ffff' is 200%--this will
    most likely generate velocities outside the midi range of 0..127.
    But that's fine since mma will adjust volumes into the valid
    range. Using very high percentages will ensure that 'ffff' notes
    are (most likely) sounded with a maximum velocity.
"""

vols = {'OFF': 0.00, 'PPPP': 0.05, 'PPP': 0.10,
        'PP': 0.25, 'P': 0.40, 'MP': 0.70,
        'M': 1.00, 'MF': 1.10, 'F': 1.30,
        'FF': 1.60, 'FFF': 1.80, 'FFFF': 2.00}

volume = vols['M']        # default global volume
nextVolume = None         # main parser sets this to the next volume
                          # when future volumes are stacked. It's used
                          # by the volume adjust to smooth out (de)crescendos.
lastVolume = volume
futureVol = []
vTRatio = .6
vMRatio = 1 - vTRatio


def adjvolume(ln):
    """ Adjust the ratio used in the volume table and track/global ratio. """

    global vols, vTRatio, vMRatio

    if not ln:
        error("Use: AdjustVolume DYN=RATIO [..]")

    notopt, ln = opt2pair(ln, toupper=True)

    if notopt:
        error("ADJUSTVOLUME: Expecting DYNAMIC=RATIO pairs")

    for v, r in ln:
        if v == 'RATIO':
            r = stof(r)

            if r < 0 or r > 100:
                error("ADJUSTVOLUME RATIO: value must be 0 to 100")

            vTRatio = r / 100
            vMRatio = 1 - vTRatio

        elif v in vols:
            vols[v] = calcVolume(r, vols[v])

        else:
            error("ADJUSTVOLUME DYNAMIC: '%s' for AdjustVolume is unknown" % v)

    if mma.MMA.debug.debug:
        dPrint("Volume Ratio: %s%% Track / %s%% Master" % (vTRatio * 100, vMRatio * 100))
        dPrint("Volume table: %s" % 
              ' '.join([ "%s=%s" % (a, int(vols[a] * 100)) for a in sorted(vols)]))


calcVolumeRePat = re.compile(
    # This is a regular expression to parse the input for the calcVolume
    # function. The details are embedded in the comments below.
    # Thanks to Johan Vromans!

    # Note, there are 2 parts to this. It's either numeric
    # or mnemonic. We use a | to join the 2 halves.
    # opt. +/- (assigned to 'pre'), numeric value ('val'), and optional % to 'post'
    # Note the '$' terminator which forces NUMERIC[%] to be end.
    r'(?P<pre>[-+])?(?P<val>[\d.]+)(?P<post>%)?$'
    + '|' +
    # mnemonic is assigned to 'mn'.
    r'(?P<mn>[OFPM]+)$',re.IGNORECASE )

def calcVolume(new, old):
    """ Calculate a new volume "new" possibly adjusting from "old". """

    m = calcVolumeRePat.match(new)
    if m is None:
        error("Unknown volume '%s'" % new)

    # Do we have a mnemonic?

    mn = m.group('mn')
    if mn:
        mn = mn.upper()
        if not mn in vols:
            error("Unknown volume mnemonic '%s'" % mn)
        return vols[mn]

    # So it must be a value.
    val  = m.group('val')
    pre  = m.group('pre')
    post = m.group('post')

    if pre and not post:
        warning("Please use '%s%s%%' if you want to %screase the volume"
                % (pre, val, "in" if pre == '+' else "de"))

    v = stof(val, "Volume expecting value, not '%s'" % val) / 100.

    if pre == '+':
        v = old * ( 1 + v )
    elif pre == '-':
        v = old * ( 1 - v )
    elif post == '%':
        v = old * ( v )

    return v

def setVolume(ln):
    """ Set master volume. """

    global volume, lastVolume, futureVol

    if len(ln) != 1:
        error("Use: Volume DYNAMIC")

    volume = calcVolume(ln[0], volume)

    futureVol = []
    if mma.MMA.debug.debug:
        dPrint("Volume: %s%%" % (100 * volume))


# The next 3 are called from the parser.

def setCresc(ln):
    """ Master Crescendo. """

    setCrescendo(1, ln)


def setDecresc(ln):
    """ Master Decrescendo (Diminuendo). """
    setCrescendo(-1, ln)


def setSwell(ln):
    """ Set a swell (cresc<>decresc). """

    global futureVol, volume, lastVolume

    lastVolume = volume

    if len(ln) == 3:            # 3 args, 1st is intial setting
        setVolume([ln[0]])
        ln = ln[1:]

    if len(ln) != 2:
        error("Swell expecting 2 or 3 args.")

    count = stoi(ln[1])
    if count < 2:
        error("Swell bar count must be 2 or greater.")

    if count % 2:
        c = (count + 1) // 2
        offset = 1
    else:
        c = count // 2
        offset = 0

    c = str(c)

    futureVol = fvolume(0, volume, [ln[0], c])
    futureVol.extend(fvolume(0, futureVol[-1],
                             [str(int(volume * 100)), c])[offset:])

    if mma.MMA.debug.debug:
        dPrint("Set Swell to: %s" % ' '.join([str(int(a * 100)) for a in futureVol]))


def setCrescendo(dir, ln):
    """ Combined (de)cresc() """

    global futureVol, volume, lastVolume

    lastVolume = volume

    if len(ln) not in (2, 3):
        error("Usage: (De)Cresc [start-Dynamic] final-Dynamic bar-count")

    if len(ln) == 3:
        setVolume([ln[0]])
        ln = ln[1:]

    futureVol = fvolume(dir, volume, ln)

    if mma.MMA.debug.debug:
        dPrint("Set (De)Cresc to: %s" % ' '.join([str(int(a * 100)) for a in futureVol]))

# Used by both the 2 funcs above and from TRACK.setCresc()


def fvolume(dir, startvol, ln):
    """ Create a list of future vols. Called by (De)Cresc. """

    # Get destination volume

    destvol = calcVolume(ln[0], startvol)

    bcount = stoi(ln[1], "Type error in bar count for (De)Cresc, '%s'" % ln[1])

    if bcount <= 0:
        error("Bar count for (De)Cresc must be postive")

    # Test to see if (de)cresc is contrary to current settings.
    # Using 'dir' of 0 will bypass this (used by SWELL).

    if dir > 0 and destvol < startvol:
        warning("Cresc volume less than current setting")

    elif dir < 0 and destvol > startvol:
        warning("Decresc volume greater than current setting")

    elif destvol == startvol:
        warning("(De)Cresc volume equal to current setting")

    if bcount > 1:
        bcount -= 1
    step = (destvol - startvol) / bcount
    volList = [startvol]

    for a in range(bcount - 1):
        startvol += step
        volList.append(startvol)

    volList.append(destvol)

    return volList


def calcMidiVolume14(s):
    """ Convert a mnemonic volume to value for MIDI channel.
        This function uses a 14 bit (0..0x3fff). See below for
        the 7 bit version.
    """

    s = s.upper()
    if s in vols:
        v = int(0x1fff * vols[s])
        if v < 0:
            v = 0
        if v > 0x3fff:
            v = 0x3fff

    else:
        v = stoi(s, "Expecting integer arg or volume mnemonic, not %s" % s)

    return v


def calcMidiVolume(s):
    """ Convert a mnemonic volume to value for MIDI channel. """

    s = s.upper()
    if s in vols:
        v = int(80 * vols[s])
        if v < 0:
            v = 0
        if v > 127:
            v = 127

    else:
        v = stoi(s, "Expecting integer arg or volume mnemonic, not %s" % s)

    return v
